//
//  BFRecordScreenController.swift
//  BFRecordScreenKit_Example
//
//  Created by 胡志强 on 2021/11/23.
//  Copyright © 2021 CocoaPods. All rights reserved.
//

import AVFAudio
import AVFoundation
import BFCommonKit
import BFMaterialKit
import BFMediaKit
import BFNetRequestKit
import BFUIKit
import Foundation
import GPUImage
import Photos
import UIKit

struct WithDrawModel {
    var type: Int // 0:拖动； 1:预览播放暂停 2: 录音结束  3: 删除录音
    var timestamp: CMTime
    var deletedVoices: [PQVoiceModel]?
    // add by ak 保存删除的字幕数据用于恢复
    var deletedTittles: [PQEditSubTitleModel]?
    //
    var deletedCameras: [PQEditVisionTrackMaterialsModel]?

}

public class BFRecordScreenController: BFBaseViewController {
    public var nextActionHandle: (() -> Void)?
    public var closeActionHandle: (() -> Void)?
    public var changeItemHandle: ((_ index: Int) -> Void)?
    // 当前录制结束的上报
    public var recordEndHandle: ((_ currentRecord: PQVoiceModel?) -> Void)?
    // 字幕按钮点击上报
    public var subTitleBtnClickHandle: ((_ isOn: Bool) -> Void)?
//    public var currentRecordId: String? // 当前录制Id

    // MARK: - 素材参数
    lazy var rsvmanager : BFRecordScreenVideoManager = {
        let m = BFRecordScreenVideoManager()
        m.assetPlayer = assetPlayer
        m.recordPlayer = recordPlayer
        m.dele = self
        return m
    }()
    
    lazy var rsimanager : BFRecordScreenImageManager = {
        let m = BFRecordScreenImageManager()
        m.recordPlayer = recordPlayer
        m.dele = self
        return m
    }()
    
    lazy var rscmanager : BFRecordScreenCameraManager = {
        let m = BFRecordScreenCameraManager()
        m.recordProgressCallBack = {[weak self] progress in
            guard let wself = self else { return }
            wself.rscurrentManager.currentAssetProgress = progress
        }
        m.hasInitCallBack = {[weak self] in
            guard let wself = self else { return }
            wself.recordBtn.isEnabled = true
        }
        
        // MARK: 摄像头结束回调
        m.recordEndCallBack = { [weak self] isSuccess, sticker in
            guard let wself = self, let sticker = sticker else { return }
            
            if isSuccess{
                wself.rscurrentManager.currentAssetProgress = sticker.timelineCMOut
                let dur = wself.rscmanager.recordItem?.videoStickers.reduce(0, { partialResult, mod in
                    (mod.timelineCMOut - mod.timelineCMIn).seconds + partialResult
                })
                wself.rscmanager.recordItem?.materialDuraion = CMTime(seconds: dur ?? 0, preferredTimescale: 1000)
//                BFLog(1, message: "camera：\(sticker.timelineCMIn.seconds) ~ \(sticker.timelineCMOut.seconds)")
            }
        }
        return m
    }()
    
    var rscurrentManager = BFRecordScreenBaseManager()
    
    // MARK: - 录制参数
    public var assets = [PHAsset]()
    
    var currItemModelIndex = 0 {
        didSet{
            if currItemModelIndex < 0 || currItemModelIndex >= itemModels.count{
                return
            }
            let itemModel = itemModels[currItemModelIndex]
            currMediaType = itemModel.mediaType ?? .Video
            isEndPlay = false

            switch itemModel.mediaType {
            case .Image:
                rscurrentManager = rsimanager
                recordBtn.setTitle("record_ly".BFLocale, for: .normal)
                progressThumV.isHidden = false
                cameraProgressThumV.isHidden = true
                progreddL.isHidden = false

            case .Video:
                rscurrentManager = rsvmanager
                recordBtn.setTitle("record_ly".BFLocale, for: .normal)
                progressThumV.isHidden = false
                cameraProgressThumV.isHidden = true
                progreddL.isHidden = false
                
            case .Camera:
                rscurrentManager = rscmanager
                recordBtn.setTitle("record_lz".BFLocale, for: .normal)
                recordBtn.isEnabled = false
                progressThumV.isHidden = true
                cameraProgressThumV.isHidden = false
                progreddL.isHidden = true
                rscmanager.cameraProgressV = cameraProgressThumV
                
            default:
                break
            }
            
            let cell = collectionView.cellForItem(at: IndexPath(item: currItemModelIndex, section: 0)) as? BFImageCoverViewCell
            if  cell != nil {
                if currMediaType == .Camera {
                    // 摄像头要旋转视频角度
                    cell!.playView.setInputRotation(GPUImageRotationMode(rawValue: 2), at: 0)
                    progressThumV.isHidden = true

                    rscmanager.rendView.removeFromSuperview()
                    cell!.playView.addSubview(rscmanager.rendView)
                    rscmanager.rendView.snp.makeConstraints { make in
                        make.edges.equalToSuperview()
                    }
                    
                } else {
                    progressThumV.isHidden = false
                    if currMediaType == .Image{
                        // 解决不能第一时间获取缩略图时的问题
                        cell!.addData()
                    }
                }
                rscurrentManager.playView = cell!.playView
            }
            
            rscurrentManager.dele = self
            rscurrentManager.progreddL = progreddL
            rscurrentManager.recordPlayer = recordPlayer
            rscurrentManager.progressThumV = progressThumV
            progressThumV.recordItem = itemModel
            rscurrentManager.recordItem = itemModel
            rscurrentManager.resetEnv()
            if currMediaType == .Video {
                (cell as? BFVideoCoverViewCell)?.rotationView()
            }
        }
    }
    
    var currMediaType : ModelType = .Video
    
    public var itemModels = [BFRecordItemModel]()
    // add by ak 当前的显示的字幕位置
    var showSubtitleIndex = -1

    var currentPlayRecordIndex: Int = -1 // >= 0 :当前正在播放的录音资源序号； -3： 刚录音完，不需要播放录音； -1:初始化阶段

    public var displayLink: CADisplayLink? // 图片素材播放控制

    public var isRecording = false { // 是否正在录音
        didSet {
            withDrawBtn.isHidden = isRecording
            changeVoiceBtn.isHidden = isRecording
            subtitleBtn.isHidden = isRecording
            soundSettingBtn.isHidden = isRecording
            recordBtn.setTitle(isRecording ? "record_pause".BFLocale : (currMediaType == .Camera ? "record_lz".BFLocale : "record_ly".BFLocale), for: .normal)
            recordBtn.backgroundColor = UIColor.hexColor(hexadecimal: "#389AFF", alpha: isRecording ? 0.6 : 1)
//            if !isRecording {
//                BFLog(1, message: "stop")
//            }
        }
    }

    var isEndPlay = false {
        didSet{
            if isEndPlay {
                
                updateRecordBtnStatus()
                
                subtitleBtn.isHidden = false
                soundSettingBtn.isHidden = false
            }else {
                
            }
        }
    }

    public var isNormalPlaying = false { // 是否正在播放
        didSet {
            withDrawBtn.isHidden = isNormalPlaying
        }
    }

    var currentAssetProgress: CMTime = .zero
    {
        didSet {
//            BFLog(1, message: "currentAssetProgress=\(currentAssetProgress.seconds)")
        }
    } // 当前素材播放的进度
    // 播放器开始播放时间
    var recordStartPlayTime: CMTime = .zero
    // 某个录音开始播放时间
    var currenStartPlayTime: CMTime = .zero
    var pauseTime: Double = 0 // 停止无操作的时间点

    var assetPlayer: AVPlayer = AVPlayer(playerItem: nil) // 原视频音频播放器
    var recordPlayer: AVPlayer = AVPlayer(playerItem: nil) // 录音音频播放器

    var hadPrepareToPlayRecord = false // 录音播放器准备
    var playView: GPUImageView? // 视频展示视图

    // MARK: 行为参数

    var movieIsProcessing = false

    // 行为记录，方便撤销
    var events = [WithDrawModel]() {
        didSet {
            withDrawBtn.isEnabled = (events.count != 0)
        }
    }

    // 是否在拖动进度条
    var isDragingProgressSlder: Bool = false

    var isStopAtRecordRange = -1

    // 保存识别出来的字幕信息,用于回放，和合成使用
    //    var saveSubTitles:[PQEditSubTitleModel] = Array.init()

    // MARK: - 录音相关

    // 定义音频的编码参数
    let recordSettings: [String: Any] = [AVSampleRateKey: 44100.0, // 声音采样率
                                         AVFormatIDKey: kAudioFormatLinearPCM, // 编码格式
                                         AVNumberOfChannelsKey: 1, // 采集音轨
                                         AVEncoderBitDepthHintKey: 16, // 位深
                                         AVEncoderAudioQualityKey: AVAudioQuality.medium.rawValue] // 音频质量

    public var haveSpeakVolume: Float = 0.15
    public var noSpeakVolume: Float = 0.30

    // 录音管理器
    var recorderManager: BFVoiceRecordManager?

    var NeoNuiAPPID: String?
    var NeoNuiToken: String?

    // MARK: - 视图参数

    var beginOnStartBtn: Bool = false
    var touchStart: CGPoint = CGPoint(x: 0, y: 0)
    var avplayerTimeObserver: NSKeyValueObservation?
    var recordPlayerTimeObserver: NSKeyValueObservation?

    var indirectionView: BFIndirectionProgressView?

    lazy var progreddL: UILabel = {
        let l = UILabel()
        l.textAlignment = .center
        l.font = UIFont.systemFont(ofSize: 13, weight: .medium)
        l.textColor = .white
        l.layer.shadowOpacity = 0.4
        l.layer.shadowColor = UIColor.black.cgColor
        l.layer.shadowOffset = CGSize(width: 0, height: 2)
        l.layer.shadowRadius = 2
        l.text = "00:00"
        return l
    }()

    public lazy var playBtn: UIButton = {
        let btn = UIButton(frame: CGRect(x: 0, y: 0, width: 150, height: 156))
        btn.setImage(imageInRecordScreenKit(by: "preview_play"), for: .normal)
        let vv = UIView(frame: CGRect(x: 0, y: 0, width: 0.5, height: 0.5))
        vv.backgroundColor = UIColor.init(white: 1, alpha: 0.3)
        btn.setImage(vv.graphicsGetImage(), for: .selected)
        btn.addTarget(self, action: #selector(playVideo(btn:)), for: .touchUpInside)
        btn.addTarget(self, action: #selector(playbtnDown(btn:)), for: .touchDown)
        btn.addTarget(self, action: #selector(playbtnCancle(btn:)), for: .touchCancel)
        btn.center = view.center
        btn.isHidden = true
        return btn
    }()

    lazy var bottomeView: UIImageView = {
        let iv = UIImageView(image: imageInRecordScreenKit(by: "bottom_shadow"))
        iv.contentMode = .scaleAspectFill
        iv.isUserInteractionEnabled = true
        // 拖曳手勢
        let pan = UIPanGestureRecognizer(target: self, action: #selector(pan(recognizer:)))
        pan.delegate = self
        pan.minimumNumberOfTouches = 1
        pan.maximumNumberOfTouches = 1
        iv.addGestureRecognizer(pan)

        return iv
    }()

    lazy var recordBtn: UIButton = {
        let btn = UIButton(type: .custom)
        btn.backgroundColor = ThemeStyleColor
        btn.setTitle("record_ly".BFLocale, for: .normal)
        btn.addCorner(corner: 6)
        btn.titleLabel?.font = UIFont.boldSystemFont(ofSize: 18)
        btn.adjustsImageWhenHighlighted = false
        btn.addTarget(self, action: #selector(startRecord), for: .touchDown)
        btn.addTarget(self, action: #selector(endRecord), for: .touchUpInside)
        btn.isEnabled = false
        return btn
    }()

    lazy var deleteRecordBtn: UIButton = {
        let btn = UIButton(type: .custom)

        btn.backgroundColor = UIColor.hexColor(hexadecimal: "#FF0000", alpha: 0.3)
        btn.setTitle("record_delete".BFLocale, for: .normal)
        btn.adjustsImageWhenHighlighted = false
        btn.titleLabel?.font = UIFont.boldSystemFont(ofSize: 18)
        btn.addCorner(corner: 6)
        btn.addTarget(self, action: #selector(deleteRecordAction), for: .touchUpInside)
        btn.eventInterval = 0.25
        btn.isHidden = true
        return btn
    }()

    lazy var withDrawBtn: UIButton = {
        let btn = UIButton(type: .custom)
        btn.setImage(imageInRecordScreenKit(by: "withdraw_n"), for: .normal)
        btn.setImage(imageInRecordScreenKit(by: "withdraw_h"), for: .highlighted)
        btn.setTitle("record_back".BFLocale, for: .normal)
        btn.setTitleColor(.white, for: .normal)
        btn.setTitleColor(.lightGray, for: .disabled)
        btn.isEnabled = false
        btn.titleLabel?.font = UIFont.systemFont(ofSize: 12)
        btn.contentVerticalAlignment = UIControl.ContentVerticalAlignment.center
        btn.addTarget(self, action: #selector(withdrawAction), for: .touchUpInside)
        return btn
    }()

    lazy var changeVoiceBtn: UIButton = {
        let btn = UIButton(type: .custom)
        btn.setImage(imageInRecordScreenKit(by: "changeVoice_n"), for: .normal)
        btn.setImage(imageInRecordScreenKit(by: "changeVoice_h"), for: .highlighted)
        btn.setTitle("变声", for: .normal)
        btn.setTitleColor(.white, for: .normal)
        btn.setTitleColor(ThemeStyleColor, for: .highlighted)
        btn.titleLabel?.font = UIFont.systemFont(ofSize: 12)
        btn.contentVerticalAlignment = UIControl.ContentVerticalAlignment.center
        btn.addTarget(self, action: #selector(changeVoiceAction), for: .touchUpInside)
        return btn
    }()

    lazy var toolV: BFIntroduceToolView = {
        let toolV = BFIntroduceToolView()
        toolV.centerY = view.centerY

        toolV.choosedToolHandle = { [weak self] tool in
            guard let wself = self else {
                return
            }
            tool.center = wself.view.center
            wself.view.addSubview(tool)
        }

        return toolV

    }()

    // 头像  add by ak
    lazy var avatarView: BFRecordAvatarView = {
        let avatarView = BFRecordAvatarView(frame: CGRect(x: 10, y: 10, width: 120, height: 120))
        avatarView.isHidden = true
        return avatarView
    }()

    // 打开摄像头
    lazy var openCameraBtn: UIButton = {
        let btn = UIButton(type: .custom)
        btn.setImage(imageInRecordScreenKit(by: "openCamera"), for: .normal)
        btn.addTarget(self, action: #selector(openCamera), for: .touchUpInside)
        return btn
    }()

    // 画笔
    lazy var drawPinBtn: UIButton = {
        let btn = UIButton(type: .custom)
        btn.setImage(imageInRecordScreenKit(by: "drawPin"), for: .normal)
        btn.addTarget(self, action: #selector(drawPin), for: .touchUpInside)
        return btn
    }()

    // 字幕设置
    lazy var subtitleBtn: UIButton = {
        let btn = UIButton(type: .custom)
        btn.setImage(imageInRecordScreenKit(by: "subtitleBtn_on"), for: .normal)
        btn.setTitle("function_subtitle".BFLocale, for: .normal)
        btn.titleLabel?.font = UIFont.systemFont(ofSize: 10)
        btn.addTarget(self, action: #selector(subTitleClick), for: .touchUpInside)
        btn.imageEdgeInsets = UIEdgeInsets(top: -20, left: 4, bottom: 0, right: -4)
        btn.titleEdgeInsets = UIEdgeInsets(top: 20, left: -40, bottom: -20, right: 0)
        return btn
    }()

    // 声音设置
    lazy var soundSettingBtn: UIButton = {
        let btn = UIButton(type: .custom)
        btn.setImage(imageInRecordScreenKit(by: "soundBtn"), for: .normal)
        btn.setTitle("function_voice".BFLocale, for: .normal)
        btn.titleLabel?.font = UIFont.systemFont(ofSize: 10)
        btn.addTarget(self, action: #selector(soundSetting), for: .touchUpInside)
        btn.imageEdgeInsets = UIEdgeInsets(top: -20, left: 4, bottom: 0, right: -4)
        btn.titleEdgeInsets = UIEdgeInsets(top: 20, left: -40, bottom: -20, right: 0)
        return btn
    }()


    // 字幕设置面板
    lazy var subtitleSettingView: BFSubtitleSettingView = {
        let subtitleSetting = BFSubtitleSettingView(frame: CGRect(x: 0, y: 0, width: cScreenWidth, height: cScreenHeigth))
        subtitleSetting.isHidden = true
        return subtitleSetting
    }()

    // 编辑字幕
    lazy var subtitleEditView: BFSubtitleEditView = {
        let subtitleEditView = BFSubtitleEditView(frame: CGRect(x: 0, y: 0, width: cScreenWidth, height: cScreenHeigth))
        subtitleEditView.isHidden = true

        return subtitleEditView
    }()

    // 显示字幕
    lazy var subtitleLabel: PQSubTitleLabel = {
        let subtitleLabel = PQSubTitleLabel()
        subtitleLabel.tag = 100_100
        subtitleLabel.textAlignment = .center
        subtitleLabel.numberOfLines = 0
        subtitleLabel.lineBreakMode = .byWordWrapping
        subtitleLabel.isUserInteractionEnabled = true
        subtitleLabel.addGestureRecognizer(UITapGestureRecognizer(target: self, action: #selector(editSubtitle)))
        return subtitleLabel
    }()

    // 音量设置
    lazy var audioSettingView: BFAudioSettingView = {
        let audioSettingView = BFAudioSettingView(frame: CGRect(x: 0, y: 0, width: cScreenWidth, height: cScreenHeigth))
        audioSettingView.isHidden = true
        audioSettingView.haveSpeakSlider.valueIsInt = true
        audioSettingView.noSpeakSlider.valueIsInt = true
        audioSettingView.callBack = { [weak self] haveSpeak, noHaveSpeak in
//            BFLog(1, message: "haveSpeak is:\(haveSpeak),noHaveSpeak is:\(noHaveSpeak)")
            guard let wself = self else { return }

            wself.haveSpeakVolume = haveSpeak / 100.0
            wself.noSpeakVolume = noHaveSpeak / 100.0
            if !wself.isNormalPlaying, !wself.isRecording {
                if wself.deleteRecordBtn.isHidden {
//                    wself.recordPlayer.volume = 0
                    wself.assetPlayer.volume = wself.noSpeakVolume
                } else {
//                    wself.recordPlayer.volume = 1.0
                    wself.assetPlayer.volume = wself.haveSpeakVolume
                }
            }
        }

        return audioSettingView
    }()

    lazy var progressThumV: BFVideoThumbProgressView = {
        let vv = BFVideoThumbProgressView(frame: CGRect(x: 0, y: 18, width: cScreenWidth, height: 50))
        self.thumViewCallBack(vv)
        vv.isHidden = true
        return vv
    }()
    
    lazy var cameraProgressThumV: BFCameraProgressView = {
        let vv = BFCameraProgressView(frame: CGRect(x: 0, y: 18, width: cScreenWidth, height: 50))
        self.thumViewCallBack(vv)
        vv.isHidden = true
        return vv
    }()

    lazy var collectionView: UICollectionView = {
        let flowLayout = UICollectionViewFlowLayout()
        flowLayout.minimumLineSpacing = 0
        flowLayout.minimumInteritemSpacing = 0
        flowLayout.scrollDirection = .horizontal
        flowLayout.itemSize = CGSize(width: view.frame.width, height: view.frame.height)
        let collectionView = UICollectionView(frame: CGRect(x: 0, y: 0, width: view.frame.width, height: view.frame.height), collectionViewLayout: flowLayout)
        collectionView.register(BFImageCoverViewCell.self, forCellWithReuseIdentifier: String(describing: BFImageCoverViewCell.self))
        collectionView.register(BFVideoCoverViewCell.self, forCellWithReuseIdentifier: String(describing: BFVideoCoverViewCell.self))
        collectionView.register(BFCameraCoverViewCell.self, forCellWithReuseIdentifier: String(describing: BFCameraCoverViewCell.self))
        collectionView.isPagingEnabled = true
        collectionView.bounces = false
        collectionView.showsVerticalScrollIndicator = false
        collectionView.showsHorizontalScrollIndicator = false
        collectionView.delegate = self
        collectionView.dataSource = self
        collectionView.backgroundColor = UIColor.clear
        if #available(iOS 11.0, *) {
            collectionView.contentInsetAdjustmentBehavior = .never
        } else {
            automaticallyAdjustsScrollViewInsets = false
        }
        return collectionView
    }()

    // debug
    lazy var neoNuiDebugLabel: UILabel = {
        let neoNuiDebugLabel = UILabel(frame: CGRect(x: 0, y: 100, width: 300, height: 24))
        neoNuiDebugLabel.backgroundColor = .orange
        neoNuiDebugLabel.textColor = .white
        neoNuiDebugLabel.textAlignment = .left
        neoNuiDebugLabel.text = "调试信息"
        return neoNuiDebugLabel

    }()

    lazy var audioQueueRecoderLabel: UILabel = {
        let audioQueueRecoderLabel = UILabel(frame: CGRect(x: 0, y: 125, width: 300, height: 24))
        audioQueueRecoderLabel.backgroundColor = .orange
        audioQueueRecoderLabel.textColor = .white
        audioQueueRecoderLabel.textAlignment = .left
        audioQueueRecoderLabel.text = "调试信息"
        return audioQueueRecoderLabel

    }()

    // MARK: - ----------------- 生命周期

    deinit {
        NotificationCenter.default.removeObserver(self)
        if avplayerTimeObserver != nil {
            assetPlayer.removeTimeObserver(avplayerTimeObserver as Any)
            avplayerTimeObserver?.invalidate()
        }
        if recordPlayerTimeObserver != nil {
            recordPlayer.removeTimeObserver(recordPlayerTimeObserver as Any)
            recordPlayerTimeObserver?.invalidate()            
        }
        
        if isRecording {
            recorderManager?.stopRecord(isCancel: true)
        }
        assetPlayer.pause()
        recordPlayer.pause()
    }

    @objc func didBecomeActive() {
        BFLog(message: "进入活跃状态")
    }

    @objc func willResignActive() {
        BFLog(message: "进入非活跃状态")
        endRecord()
    }

    /// 线路切换
    /// - Parameter nofify: <#nofify description#>
    @objc func routeChangeNofify(nofify: Notification) {
        let routeChangeDic = nofify.userInfo
        let routeChangeReason: AVAudioSession.RouteChangeReason? = AVAudioSession.RouteChangeReason(rawValue: UInt((routeChangeDic?[AVAudioSessionRouteChangeReasonKey] as? Int) ?? 0))
        if routeChangeReason == .oldDeviceUnavailable {
            let previousRoute = routeChangeDic?[AVAudioSessionRouteChangePreviousRouteKey] as? AVAudioSessionRouteDescription
            if previousRoute?.outputs.first?.portType == .headphones {
                playVideo(btn: playBtn)
            }
        }
    }

    override public func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)
        navigationController?.isNavigationBarHidden = true
        hiddenNavigation()
        UIApplication.shared.isIdleTimerDisabled = true
    }
    
    public override func viewDidAppear(_ animated: Bool) {
        super.viewDidAppear(animated)
//        endScrollItem(page: max(currItemModelIndex, 0))
        // 设置默认值
        setSubtitleStyle(settingModel: subtitleSettingView.subtitle.setting)
    }

    override public func viewWillDisappear(_ animated: Bool) {
        super.viewWillDisappear(animated)
        UIApplication.shared.keyWindow?.isUserInteractionEnabled = true
        pause()
        UIApplication.shared.isIdleTimerDisabled = false
    }

    override public func viewDidLoad() {
        super.viewDidLoad()
        _ = disablePopGesture()

        // 进入活跃状态
        PQNotification.addObserver(self, selector: #selector(didBecomeActive), name: UIApplication.didBecomeActiveNotification, object: nil)
        PQNotification.addObserver(self, selector: #selector(routeChangeNofify(nofify:)), name: AVAudioSession.routeChangeNotification, object: nil)
        // 进入非活跃状态
        PQNotification.addObserver(self, selector: #selector(willResignActive), name: UIApplication.willResignActiveNotification, object: nil)

        _ = checkStatus()
        // mdf by ak 切换外放和请求权限分开 否则会导致首次安装时不能执行切换操作
        let options: AVAudioSession.CategoryOptions = [.defaultToSpeaker, .allowBluetooth]
        try? AVAudioSession.sharedInstance().setCategory(.playAndRecord, options: options)

        // add by ak 取 nsl token
        BFRecordScreenViewModel.getNlsAccessToken { [weak self] token, appkey in
            guard let wself = self else { return }
            
            BFLog(message: "nls appkey is \(appkey), token is \(token)")
            wself.NeoNuiAPPID = appkey
            wself.NeoNuiToken = token
        }
        
        initlizeRecordManager()
        
        view.backgroundColor = .black
        view.addSubview(collectionView)
                
        view.addSubview(bottomeView)
//        view.addSubview(subtitleLabel)
        view.addSubview(playBtn)
        view.addSubview(avatarView)
//        view.addSubview(openCameraBtn)
//        view.addSubview(drawPinBtn)
        view.addSubview(subtitleBtn)
        view.addSubview(soundSettingBtn)
        view.addSubview(subtitleSettingView)

        view.addSubview(audioSettingView)

        bottomeView.addSubview(progreddL)
        //        view.addSubview(toolV)
        bottomeView.addSubview(recordBtn)
        bottomeView.addSubview(deleteRecordBtn)
        bottomeView.addSubview(withDrawBtn)
        //        bottomeView.addSubview(changeVoiceBtn)
        bottomeView.addSubview(progressThumV)
        bottomeView.addSubview(cameraProgressThumV)

        avatarView.recordEndCallBack = { _, materialsModel in
            BFLog(message: "新录制完成::::\(materialsModel?.locationPath ?? "")")
        }


        // 字幕设置回调
        subtitleSettingView.subtitleSettingCallBack = { [weak self] subtitileModel in

            if let wself = self{
                
                //1，设置字幕样式
                wself.setSubtitleStyle(settingModel: subtitileModel.setting)
                
                //2，显示设置文字，不显示设置为空
                if(subtitileModel.setting.subtitleIsShow){
                    wself.updateSubtitle(time: wself.currentAssetProgress)
                }else{
                    wself.subtitleLabel.text = ""
                }
        
                // mdf by ak 这里是设置字幕开关回调
                if wself.subTitleBtnClickHandle != nil {
                    wself.subTitleBtnClickHandle!(subtitileModel.setting.subtitleIsShow)
                }
            }
         
        }
        // 编辑字幕完成
        subtitleEditView.editSubtitleDone = { [weak self] newtext, index in
            guard let wself = self else { return }

            // 更新缓存数据
            if index < (wself.itemModels[wself.currItemModelIndex ].titleStickers.count ) {
                if newtext.count == 0 {
                    // 删除数据
                    BFLog(message: "清空字幕操作要删除原字幕 sticker 数据")
                    wself.itemModels[wself.currItemModelIndex ].titleStickers.remove(at: index)
                    wself.subtitleLabel.text = ""
                } else {
                    // 更新数据
                    wself.itemModels[wself.currItemModelIndex ].titleStickers[index].text = newtext
                    wself.subtitleLabel.text = newtext
                }
            }

            wself.setSubtitleStyle(settingModel: wself.subtitleSettingView.subtitle.setting)
        }

        layoutsubview()
        view.addSubview(subtitleEditView)
        
//        DispatchQueue.main.asyncAfter(deadline: .now() + 0.25) {[weak self] in
//            guard let wself = self else { return }
//            
//            wself.endScrollItem(page: 0)
//        }

        if !SWNetRequest.isNetReachabled() {
            cShowHUB(superView: view, msg: "网络不佳，字幕可能无法生成")
        }
        if PQENVUtil.shared.channel == "Development" {
            view.addSubview(neoNuiDebugLabel)
            view.addSubview(audioQueueRecoderLabel)
        }
    }
    
    func layoutsubview() {
        bottomeView.snp.makeConstraints { make in
            make.left.bottom.right.equalToSuperview()
            make.height.equalTo(adapterWidth(width: 180))
        }

        progreddL.snp.makeConstraints { make in
            make.width.equalTo(100)
            make.centerX.equalToSuperview()
            make.top.equalToSuperview().offset(-5)
            make.height.equalTo(18)
        }

        withDrawBtn.snp.makeConstraints { make in
            make.left.equalToSuperview()
            make.width.height.equalTo(65)
            make.top.equalTo(88)
        }
        //        changeVoiceBtn.snp.makeConstraints { make in
        //            make.right.equalToSuperview()
        //            make.top.width.height.equalTo(withDrawBtn)
        //        }

        recordBtn.snp.makeConstraints { make in
            make.left.equalTo(withDrawBtn.snp.right)
            make.right.equalTo(-65)
            make.height.equalTo(42)
            make.top.equalTo(withDrawBtn).offset(6)
        }
        deleteRecordBtn.snp.makeConstraints { make in
            make.left.equalTo(withDrawBtn.snp.right)
            make.right.equalTo(-65)
            make.height.equalTo(42)
            make.top.equalTo(withDrawBtn).offset(6)
        }

        //        openCameraBtn.snp.makeConstraints { make in
        //            make.right.equalToSuperview().offset(-12)
        //            make.top.equalToSuperview().offset(98)
        //            make.width.equalTo(80)
        //            make.height.equalTo(124)
        //        }
        //
        //        drawPinBtn.snp.makeConstraints { make in
        //            make.right.equalTo(openCameraBtn)
        //            make.top.equalTo(openCameraBtn.snp.bottom).offset(18)
        //            make.width.equalTo(80)
        //            make.height.equalTo(124)
        //        }

        soundSettingBtn.snp.makeConstraints { make in
            make.right.equalToSuperview().offset(-8)
            make.top.equalToSuperview().offset(98)
            make.width.equalTo(48)
            make.height.equalTo(62)
        }

        subtitleBtn.snp.makeConstraints { make in
            make.right.width.height.equalTo(soundSettingBtn)
            make.top.equalTo(soundSettingBtn.snp.bottom).offset(18)
        }

        withDrawBtn.imageEdgeInsets = UIEdgeInsets(top: -withDrawBtn.imageView!.height, left: 0, bottom: 0, right: -withDrawBtn.titleLabel!.width)
        withDrawBtn.titleEdgeInsets = UIEdgeInsets(top: withDrawBtn.titleLabel!.height + 2, left: -withDrawBtn.imageView!.width, bottom: 0, right: 0)

        changeVoiceBtn.imageEdgeInsets = UIEdgeInsets(top: -changeVoiceBtn.imageView!.height - 2, left: 0, bottom: 0, right: -changeVoiceBtn.titleLabel!.width)
        changeVoiceBtn.titleEdgeInsets = UIEdgeInsets(top: changeVoiceBtn.titleLabel!.height + 2, left: -changeVoiceBtn.imageView!.width, bottom: 0, right: 0)
    }
    // MARK: - 录音机字幕回调相关
    func initlizeRecordManager() {
        recorderManager = BFVoiceRecordManager()
        // 录音进度
        recorderManager?.recorderProgrossHandle = { [weak self] progress in
            guard let wself = self else { return }
            wself.drawProgressIndication(progress: (progress.isNaN || progress.isInfinite) ? 0 : progress)
        }

        // MARK: 录音字幕回调
        recorderManager?.subtitleRecordHandle = { [weak self] eventCode, recordId, asrResult, audioFilePath, _ in
//            BFLog(1, message: "eventcode:\(eventCode), recordid:\(String(describing: recordId)), asr:\(String(describing: asrResult))")

            DispatchQueue.global().async { [weak self] in
                guard let wself = self else { return }

                let newSubtitle = PQEditSubTitleModel()
                newSubtitle.recordId = recordId

                if asrResult == nil || (asrResult?.count ?? 0) <= 0 {
//                    BFLog(1, message: "onNuiEventCallback 识别结果为空？？？？不能生成字幕数据")
                    return
                }
                // 只有为24时才有字幕返回
                if eventCode != 24 || recordId == nil{
//                    BFLog(1, message: "onNuiEventCallback eventCode != 24 || recordId == nil")

                    return
                }
                
                let dicResult: [String: Any]? = jsonStringToDictionary(asrResult!)
                let payload = dicResult?["payload"] as? [String: Any]
//                let header = dicResult?["header"] as? [String: Any]
//                BFLog(1, message: "onNuiEventCallback event 识别结果:) \(payload?["result"] ?? "") ,taskId:\((header?["task_id"] as? String) ?? "taskId"), 识别时间：\(((payload?["begin_time"]) as? Int) ?? 0) ~ \(((payload?["time"]) as? Int) ?? 0) startTime:\(wself.recorderManager?.voiceModel?.startCMTime.seconds ?? 0.0)")
                // 1，保存字幕数据 begin_time是开始出现文字的时间，time 是结束文字出现的时间 单位都为毫秒，都是相对于录制音频数据整段时间。self.recorderManager.voiceModel?.startCMTime.seconds 为开始的录制的时间，开始和结束都要加上这个时差
                // 这里加300ms 是因为返回结果为了切到字，时长提前一些时间，具体时间官方没说和原音频有关系。这里我们先延后300ms 单位：毫秒。
                var tempVoice: PQVoiceModel?
                var tempItem: BFRecordItemModel?
                // 1:先通过titleTaskId来找是否存在录音
                wself.itemModels.enumerated().forEach { index, item in
                    if tempVoice == nil {
                        tempVoice = item.voiceStickers.first { voice in
                            voice.recordId == recordId
                        }
                        if tempVoice != nil {
                            tempItem = item
                        }
//                        BFLog(3, message: "字幕回调-找到录音文件:taskID=\(recordId ),audioFilePath=\(audioFilePath ?? ""),index=\(index)")
                    }
                }
                // 2:如果通过titleTaskId没找到录音文件,则通过
                if tempVoice == nil && (wself.isRecording ) && wself.itemModels[wself.currItemModelIndex ].voiceStickers.last?.recordId == nil {
//                    BFLog(3, message: "字幕回调-如果通过titleTaskId没找到录音文件:taskID=\(recordId ),audioFilePath=\(audioFilePath ?? "")")
                    tempVoice = wself.itemModels[wself.currItemModelIndex ].voiceStickers.last
                }
                // 3:如果通过titleTaskId跟audioFilePath都没找到录音文件,则默认为recorderManager?.voiceModel
                guard let currentVoice = (tempVoice ?? wself.recorderManager?.voiceModel) ?? wself.itemModels[wself.currItemModelIndex ].voiceStickers.last else {
//                    BFLog(3, message: "字幕回调-最终没找到录音文件:taskID=\(recordId ?? "recordId" ),audioFilePath=\(audioFilePath ?? "")")
                    return
                }
                
                if tempItem == nil {
                    tempItem = wself.itemModels[wself.currItemModelIndex]
                }

                newSubtitle.timelineIn = currentVoice.startCMTime + CMTime(seconds: Float64((((payload?["begin_time"]) as? Int) ?? 0) + 300) / 1000.0, preferredTimescale: 1000)
                newSubtitle.timelineOut = currentVoice.startCMTime + CMTime(seconds: Float64((((payload?["time"]) as? Int) ?? 0) - 10) / 1000.0, preferredTimescale: 1000)
                newSubtitle.audioFilePath = currentVoice.wavFilePath
                BFLog(3, message: "字幕回调-字幕返回begin:\(((payload?["begin_time"]) as? Int) ?? 0),timelineIn:\(newSubtitle.timelineIn.seconds), end:\(((payload?["time"]) as? Int) ?? 0),timelineOut:\(newSubtitle.timelineOut.seconds)")
                if (newSubtitle.timelineIn - currentVoice.startCMTime).seconds > 0.1 {
//                    BFLog(1, message: "卡在录音尾巴上了1")
                    newSubtitle.timelineIn = newSubtitle.timelineIn - CMTime(seconds: 0.1, preferredTimescale: 1000)
                }
                let showText = ((payload?["result"]) as? String) ?? ""
                newSubtitle.text = showText
//                newSubtitle.audioFilePath = audioFilePath ?? ""
//                BFLog(1, message: "添加字幕数据 timelineIn \(newSubtitle.timelineIn.seconds) timelineOut \(newSubtitle.timelineOut.seconds) text: \(newSubtitle.text)")
                newSubtitle.setting = wself.subtitleSettingView.subtitle.setting
                tempItem?.titleStickers.append(newSubtitle)
            }
        }

        // MARK: 录音结束
        recorderManager?.endRecordHandle = { [weak self, weak recorderManager] voideModel, _ in
            
            guard let wself = self else { return }
            
            if let model = voideModel, FileManager.default.fileExists(atPath: model.wavFilePath ?? "") {
                // 加入到语音数组里

                model.endCMTime = wself.currentAssetProgress
//                BFLog(1, message: "录音地址：\((model.wavFilePath ?? "").replacingOccurrences(of: documensDirectory, with: ""))")
                BFLog(1, message: "   文件时长record:\(wself.rscmanager.recordItem?.voiceStickers.count ?? 0),   \(model.startCMTime.seconds) -- \(model.endCMTime.seconds), dur: \(model.endCMTime.seconds - model.startCMTime.seconds)")

                if model.endCMTime.seconds < model.startCMTime.seconds{
                    BFLog(1, message: "error")
                }
                /// 注：录音机回调的录音时长大于一秒，而业务逻辑计算的会小于一秒
                if (model.endCMTime.seconds - model.startCMTime.seconds) < 1 {
                    // 取消录制
                    wself.recordManagerCancelRecord(voiceModel: model)
                    wself.rscurrentManager.recordFinished(isSuccess: false)
                    return
                }
                wself.rscurrentManager.recordFinished(isSuccess:true)
                // ********** 开始处理冲突的录制部分
                let newRange = CMTimeRange(start: model.startCMTime, end: model.endCMTime)

                var deletedVoices = [PQVoiceModel]()
                // 要删除的字幕
                var deletedTitlesTemp = [PQEditSubTitleModel]()

                // 查找要删除的音频和字幕数据
                for  m in wself.itemModels[wself.currItemModelIndex].voiceStickers {
                    let originRange = CMTimeRange(start: m.startCMTime, end: CMTime(seconds: m.endCMTime.seconds - 0.02, preferredTimescale: 1000))
                    if CMTimeRangeGetIntersection(originRange, otherRange: newRange).duration.seconds > 0 {
                        deletedVoices.append(m)
                        deletedTitlesTemp += wself.deleteTitles(voiceModel: m)
                        continue
                    }
                }
                // 删除冲突的音频
                deletedVoices.forEach { m in
                    wself.itemModels[wself.currItemModelIndex].voiceStickers.removeAll { tempM in
                        m.wavFilePath == tempM.wavFilePath
                    }
                }

                wself.itemModels[wself.currItemModelIndex].voiceStickers.append(model)
                // 录制结束回调
                wself.recordEndHandle?(model)
                // 如果是图片素材同时有需要删除的录音时需要调整录音文件开始结束时间
                // warning: 图片录制的时候应该只能在结尾处录制
                if wself.currMediaType == .Image {
                    if deletedVoices.count > 0 {
                        // 如果是图片先排序在计算区间
                        wself.itemModels[wself.currItemModelIndex].voiceStickers = wself.itemModels[wself.currItemModelIndex].voiceStickers.sorted { voice1, voice2 in
                            voice1.startCMTime.seconds < voice2.startCMTime.seconds
                        }
                        for (index, item) in wself.itemModels[wself.currItemModelIndex].voiceStickers.enumerated() {
                            if index > 0 {
                                // 注：开始时间减去duration or 等一前一段录音的结束时间
                                let tempDuration = item.endCMTime.seconds - item.startCMTime.seconds
                                item.startCMTime = CMTime(seconds: wself.itemModels[wself.currItemModelIndex].voiceStickers[index - 1].endCMTime.seconds, preferredTimescale: 1000)
                                item.endCMTime = CMTime(seconds: item.startCMTime.seconds + tempDuration, preferredTimescale: 1000)
                            }
                            BFLog(1, message: "录制结束重新排序录音文件：\(index)-\(item.wavFilePath ?? "")-\(item.startCMTime.seconds)-\(item.endCMTime.seconds)-\(item.endCMTime.seconds - item.startCMTime.seconds)")
                        }
                    }
                }
                // ***********处理冲突的录音部分资源 end

                // 添加撤销记录点
                var event = WithDrawModel(type: 2, timestamp: model.startCMTime)
                event.deletedVoices = deletedVoices
                event.deletedTittles = deletedTitlesTemp
                wself.events.append(event)

                if wself.currMediaType != .Video {
                    var duration: CMTime = .zero
                    if wself.currMediaType == .Image{
                        wself.itemModels[wself.currItemModelIndex].voiceStickers.forEach { temp in
                            BFLog(1, message: "录制结束-最终：\(temp.wavFilePath ?? "")-\(temp.startCMTime.seconds)-\(temp.endCMTime.seconds)-\(temp.endCMTime.seconds - temp.startCMTime.seconds)")
                            temp.duration = "\(temp.endCMTime.seconds - temp.startCMTime.seconds)"
                            duration = duration + temp.endCMTime - temp.startCMTime
                        }
                        wself.itemModels[wself.currItemModelIndex].materialDuraion = duration
                    }
                    
                    wself.isEndPlay = true
                    // 录制结束显示播放按钮
                    if wself.currMediaType == .Camera {
                        wself.playBtn.isSelected = true
                        wself.playBtn.isHidden = true
                    }else {
                        wself.playBtn.isSelected = wself.itemModels[wself.currItemModelIndex].voiceStickers.count <= 0
                        wself.playBtn.isHidden = wself.playBtn.isSelected
                    }
                }
                DispatchQueue.main.async { [weak self] in
                    guard let wself = self else { return }

                    // 录音完，重绘撤销按钮，更新录音按钮，
                    wself.changeWithDrawBtnLayout(2)
                    // 注：在录制结束时矫正当前位置，避免跟指针无法对其
                    wself.indirectionView?.resetCurrentItem(start: model.startCMTime.seconds, end: model.endCMTime.seconds)
                    if deletedVoices.count > 0 {
                        /// 重绘录音进度视图
                        wself.resetAllIndirectionView()
                    }
                    // 矫正进度
                    wself.resetCurrentProgress()
                    wself.deleteRecordBtn.isHidden = true
                    wself.recordBtn.isHidden = (wself.currMediaType != .Video && (wself.isEndPlay )) ? false : (wself.isEndPlay )
                }
                wself.currentPlayRecordIndex = -3 // 刚录音完，不需要播放录音
                BFLog(3, message: "重置播放index-\(#function) = \(wself.currentPlayRecordIndex)")
                recorderManager?.voiceModel = nil
            } else {
                wself.rscurrentManager.recordFinished(isSuccess: false)
                BFLog(2, message: "数据出错！！！！\(voideModel?.wavFilePath ?? "")")
            }
        }
        recorderManager?.cancelRecordHandle = { [weak self] voiceModel in
            guard let wself = self else { return }
            // 取消录制
            wself.recordManagerCancelRecord(voiceModel: voiceModel)
        }

        recorderManager?.NeoNuiDebugHandle = { [weak self] msg,isShow in
            
            guard let wself = self else { return }
            
            wself.neoNuiDebugLabel.text = "字幕服务:\(msg ?? "")"
            if(isShow){
                cShowHUB(superView: wself.view, msg: msg)
            }
        
        }

        recorderManager?.AudioQueueRecoderDebugHandle = { [weak self] msg,isShow in
            guard let wself = self else { return }
            wself.audioQueueRecoderLabel.text = "录音机:\(msg ?? "")"
            if(isShow){
                cShowHUB(superView: wself.view, msg: msg)
            }
        }

    }

    @objc func editSubtitle() {
        BFLog(message: "编辑字幕 index:\(showSubtitleIndex)")
        if !isNormalPlaying, subtitleLabel.text?.count != 0 {
            subtitleEditView.isHidden = false
            subtitleEditView.textView.becomeFirstResponder()
            subtitleEditView.setNewText(text: subtitleLabel.text ?? "", index: showSubtitleIndex)
        } else {
            BFLog(message: "播放状态,内容为空时不可以编辑字幕")
        }
    }

    /// 更新字幕，在回放时使用
    /// - Parameter time: 当前播放的进度
    func updateSubtitle(time: CMTime) {
//        BFLog(1, message: "字幕currTime is \(CMTimeGetSeconds(time))")
        if isRecording || !subtitleSettingView.subtitle.setting.subtitleIsShow || currentPlayRecordIndex == -3 {
            // currentPlayRecordIndex == -3 代表刚录完音，不要找字幕，会因为小幅度回退seek导致上一个字幕出现
            if subtitleLabel.text?.count ?? 0 > 0 {
                subtitleLabel.text = ""
                subtitleLabel.backgroundColor = UIColor.clear
            }
            return
        }
        
        if let recordItem = rscurrentManager.recordItem {
            
            if showSubtitleIndex > -1, showSubtitleIndex < recordItem.titleStickers.count {
                let currShowSubtitle = recordItem.titleStickers[showSubtitleIndex]
                
                if CMTimeCompare(currShowSubtitle.timelineIn, time) <= 0, CMTimeCompare(currShowSubtitle.timelineOut, time) > 0 {
                    // 已存在字幕
                    if subtitleLabel.text != currShowSubtitle.text {
                        subtitleLabel.text = currShowSubtitle.text
                        setSubtitleStyle(settingModel: subtitleSettingView.subtitle.setting)
                    }
                    return
                }
                showSubtitleIndex = -1
            }
            
            var findShowSubtitle: PQEditSubTitleModel?
            for (index, subtitle) in recordItem.titleStickers.enumerated() {
                if CMTimeCompare(subtitle.timelineIn, time) <= 0, CMTimeCompare(subtitle.timelineOut, time) > 0, subtitle.audioFilePath.count > 0 { //  audioFilePath.count 这个条件是确保这个字幕有对应录音
                    findShowSubtitle = subtitle
                    //                BFLog(1, message: "找到要显示的字幕 in \((findShowSubtitle?.timelineIn ?? .zero).seconds) out \((findShowSubtitle?.timelineOut ?? .zero).seconds) text:\(findShowSubtitle?.text ?? "") currTime is \(CMTimeGetSeconds(time))")
                    showSubtitleIndex = index
                    break
                }
            }
            
            if findShowSubtitle != nil, subtitleLabel.text != findShowSubtitle!.text {
                subtitleLabel.text = findShowSubtitle!.text
                setSubtitleStyle(settingModel: subtitleSettingView.subtitle.setting)
                
            } else {
                if subtitleLabel.text?.count ?? 0 > 0 {
                    subtitleLabel.text = ""
                    subtitleLabel.backgroundColor = UIColor.clear
                }
            }
        }

    }

    /// 设置字幕样式和位置
    /// - Parameter settingModel: 样式model
    func setSubtitleStyle(settingModel: BFSubTitileSettingModel) {
        subtitleBtn.setImage(imageInRecordScreenKit(by: settingModel.subtitleIsShow ? "subtitleBtn_on" : "subtitleBtn_off"), for: .normal)
        if settingModel.subtitleIsShow {
            // 设置样式
            subtitleLabel.strokeColor = settingModel.strokeColor
            subtitleLabel.strokeWidth = settingModel.strokeWidth * cScreenWidth / 1080
            subtitleLabel.font = UIFont.boldSystemFont(ofSize: CGFloat(settingModel.subtitleSize) * cScreenWidth / 1080)
            subtitleLabel.textColor = settingModel.fontColor

            let leftPoint: CGFloat = 37

            let height = sizeTextFits(attributedText: nil, text: subtitleLabel.text ?? "", numberOfLines: 0, font: subtitleLabel.font, maxSize: CGSize(width: cScreenWidth * 0.75, height: CGFloat.greatestFiniteMagnitude)).height

            subtitleLabel.backgroundColor = subtitleSettingView.subtitle.setting.backgroundColor
            // 下 //设置位置
            if subtitleSettingView.subtitle.setting.subtitlePoint == 0 {
                subtitleLabel.frame = CGRect(x: leftPoint, y: cScreenHeigth * 0.70 - height, width: cScreenWidth - 37 * 2, height: height)
            } else if subtitleSettingView.subtitle.setting.subtitlePoint == 1 { // 中
                subtitleLabel.frame = CGRect(x: leftPoint, y: cScreenHeigth / 2.0 - height / 2.0, width: cScreenWidth - 37 * 2, height: height)
            } else { // 上
                subtitleLabel.frame = CGRect(x: leftPoint, y: cScreenHeigth * 0.12, width: cScreenWidth - 37 * 2, height: height)
            }
          
        }
    }

    //    public override func viewWillLayoutSubviews() {
    //        super.viewWillLayoutSubviews()
    //
    //    }

    // MARK: - 按钮事件响应

    override public func backBtnClick() {
        pause()
    }

    // 触发拖曳手势后，执行的动作
    @objc func pan(recognizer: UIPanGestureRecognizer) {
        // 设置 UIView 新的位置
        if !checkStatus(show: false) {
            return
        }
        switch recognizer.state {
        case .began:
            touchStart = recognizer.location(in: bottomeView)
            beginOnStartBtn = (beginOnStartBtn || recordBtn.frame.contains(touchStart))
            BFLog(1, message: "beginOnStartBtn:\(beginOnStartBtn)")
        case .changed:
            if beginOnStartBtn == true {
                let nowPoint = recognizer.location(in: bottomeView)
//                BFLog(1, message: "nowPoint x: \(nowPoint.x) y:\(nowPoint.y)")
                if recordBtn.frame.contains(nowPoint) {
                } else {}
            }

        case .ended:
//            BFLog(1, message: "移动结束 \(beginOnStartBtn)")
            if beginOnStartBtn == true {
                beginOnStartBtn = false
                endRecord()
                //                if recordBtn.frame.contains(point) {
                //                    // 结束录制
                //                } else {
                //                    cancleRecord()
                //                }
            }

        default: break
//            BFLog(1, message: "pan default")
        }
    }

    @objc func openCamera() {
        if avatarView.isHidden {
            avatarView.isHidden = false
            avatarView.openCamera()
        } else {
            avatarView.isHidden = true
            avatarView.closeCamera()
        }
    }

    @objc func drawPin() {}

    // 字幕
    @objc func subTitleClick() {
        BFLog(message: "subTitle Click ")
        subtitleSettingView.isHidden = !subtitleSettingView.isHidden

        updateSubtitle(time: currentAssetProgress)
    }

    // 声音设置
    @objc func soundSetting() {
        BFLog(message: "设置声音")

        audioSettingView.isHidden = false
    }

    /// 删除指定段落的所有字幕 数据
    /// - Parameter voiceModel: 删除的音频数据
    func deleteTitles(voiceModel: PQVoiceModel) -> [PQEditSubTitleModel] {
        BFLog(message: "itemModels[currItemModelIndex].titleStickers  删除前：\(itemModels[currItemModelIndex].titleStickers.count) model.startCMTime.seconds: \(voiceModel.startCMTime.seconds) model.end: \(voiceModel.endCMTime.seconds)")
        var deleteTemp = [PQEditSubTitleModel]()

        // 从原数组中删除
        itemModels[currItemModelIndex].titleStickers.removeAll { m in
            let su = (m.recordId == voiceModel.recordId)
            if su {
                deleteTemp.append(m)
            }
            return su
        }

        BFLog(message: "itemModels[currItemModelIndex].titleStickers  删除后：\(itemModels[currItemModelIndex].titleStickers.count)")

        // 清空字幕UI
        subtitleLabel.text = ""
        subtitleLabel.backgroundColor = UIColor.clear

        return deleteTemp
    }

    @objc func deleteRecordAction() {
    
        if !isDragingProgressSlder, isStopAtRecordRange != -1, isStopAtRecordRange < itemModels[currItemModelIndex].voiceStickers.count {
            let model = itemModels[currItemModelIndex].voiceStickers[isStopAtRecordRange]
            itemModels[currItemModelIndex].voiceStickers.remove(at: isStopAtRecordRange)
            indirectionView?.deleteItem(index: isStopAtRecordRange)
            var event = WithDrawModel(type: 3, timestamp: currentAssetProgress, deletedVoices: [model])
            event.deletedTittles = deleteTitles(voiceModel: model)
            events.append(event)

            rscurrentManager.deleteRecord(at: currentAssetProgress)
            
            if currMediaType != .Video{
                // 判断是否无录音了
                if itemModels[currItemModelIndex].materialDuraion == .zero {
                    playBtn.isSelected = true
                    playBtn.isHidden = true
                }
            }

            searchStopAtRecordRange()
            changeWithDrawBtnLayout(3)
        }
    }

    @objc func startRecord() {
        //        BFLog(1, message: "start \(UIControl.Event.touchDown)")
        if !checkStatus(show: true) {
            BFLog(message: "没有录音权限不能录制")
            return
        }

        // 停止进度条滚动
        if isDragingProgressSlder {
            return
        }
        
        //add by ak 如果是云端未下载的素材在没有下载完并生成封面时 不能进行录制。
        if(progressThumV.recordItem?.thumbImgs.count == 0 && itemModels[currItemModelIndex].mediaType != .Camera){
            cShowHUB(superView:view, msg: "option_loading".BFLocale)
            return
        }

//        progressThumV.progressView.isUserInteractionEnabled = false
//        collectionView.isScrollEnabled = false
        // 开始录制时清除屏幕上的字幕
        subtitleLabel.text = ""
        subtitleLabel.backgroundColor = UIColor.clear

        pause()
        if recorderManager == nil {
            BFLog(message: "录音机初始化错误！！！")
            return
        }

        DispatchQueue.global().async {[weak self] in
            guard let wself = self else { return }
            
            wself.recorderManager?.cancelTitleService()
            let model = PQVoiceModel()
            model.startCMTime = wself.currentAssetProgress
            model.currIndex = wself.currItemModelIndex
    //        model.recordId = getUniqueId(desc: "recordId")
            model.volume = 100
            wself.recorderManager?.voiceModel = model
            BFLog(3, message: "开始录制-开始：currentAssetProgress=\(wself.currentAssetProgress.seconds),cuInde=\(wself.currItemModelIndex),\(model)")
            wself.recorderManager?.startRecord()
            wself.recorderManager?.audioRecorder?.startNeoNui(wself.NeoNuiToken ?? "", appid: wself.NeoNuiAPPID ?? "")
        }
        // 开始时间
        beginOnStartBtn = true
        isRecording = true

        UIApplication.shared.keyWindow?.isUserInteractionEnabled = false
        AudioServicesPlaySystemSound(1520)

        
        if !avatarView.isHidden {
            avatarView.beginRecord()
        }

        rscurrentManager.startRecord()
        // 录制中不显示播放按钮
        playBtn.isSelected = true
        playBtn.isHidden = true
        let cell = collectionView.cellForItem(at: IndexPath(item: currItemModelIndex, section: 0)) as? BFImageCoverViewCell
        if cell?.contentView.viewWithTag(100_100)?.superview == nil {
            cell?.contentView.addSubview(subtitleLabel)
        }
    }

    @objc func endRecord() {
        // 存储录音
        if !isRecording {
            return
        }
        if currMediaType == .Video {
            pause()
        }
        beginOnStartBtn = false

        BFLog(1, message: "停止录音- \(currentAssetProgress.seconds)")
        recordBtn.isEnabled = false
        DispatchQueue.main.asyncAfter(deadline: .now() + 0.25) {[weak self] in
            guard let wself = self else { return }
            
            wself.recordBtn.isEnabled = true
        }

        rscurrentManager.endRecord()
        
        isRecording = false
//        progressThumV.progressView.isUserInteractionEnabled = true
//        collectionView.isScrollEnabled = true

        UIApplication.shared.keyWindow?.isUserInteractionEnabled = true

        if currentAssetProgress.seconds - (recorderManager?.voiceModel?.startCMTime.seconds ?? 0) > 1.0 {
        } else {
//            recorderManager?.stopRecord(isCancel: true)
            changeProgress(changCMTime: recorderManager?.voiceModel?.startCMTime ?? .zero)
        }
        recorderManager?.stopRecord(isCancel: false)

        /*
         if currentAssetProgress.seconds - (recorderManager?.voiceModel?.startCMTime.seconds ?? 0 ) >= 1.0 {
             recorderManager?.stopRecord(isCancel: false)
             changeWithDrawBtnLayout(2)
         }else{
             recorderManager?.stopRecord(isCancel: true)
             changeProgress(progress: Float(recorderManager?.voiceModel?.startCMTime.seconds ?? 0))
         }
         */
        if !avatarView.isHidden {
            avatarView.endRecord()
        }
    }

    @objc func cancleRecord() {
        isRecording = false
        recorderManager?.stopRecord(isCancel: true)
//        progressThumV.progressView.isUserInteractionEnabled = true
//        collectionView.isScrollEnabled = true
        UIApplication.shared.keyWindow?.isUserInteractionEnabled = true
        pause()
    }

    /// 不足一秒，主动取消录制
    /// - Parameter voiceModel: <#voiceModel description#>
    @objc func recordManagerCancelRecord(voiceModel: PQVoiceModel?) {
        // 录制结束回调
        recordEndHandle?(voiceModel)
        BFLog(3, message: "开始录制-取消：currentAssetProgress=\(currentAssetProgress.seconds),cuInde=\(currItemModelIndex),currIndex=\(voiceModel?.currIndex ?? 0),\(String(describing: voiceModel)),\(String(describing: recorderManager?.voiceModel))")
        if voiceModel != nil, currentAssetProgress.seconds - (recorderManager?.voiceModel?.startCMTime.seconds ?? 0) < 1.0 {
            cShowHUB(superView: nil, msg: "shoot_tips_least".BFLocale)
            recorderManager?.cancelTitleService()
        }
        // 删除文件
        recorderManager?.deleteFile(outfile: voiceModel?.wavFilePath ?? "")
        var subtitleCount = itemModels[currItemModelIndex].titleStickers.count
        BFLog(2, message: "删除\(voiceModel?.wavFilePath ?? "")对应的字幕 前 count\(subtitleCount)")
        if subtitleCount > 0 {
            for title in itemModels[currItemModelIndex].titleStickers {
                if title.recordId == voiceModel?.recordId ?? "none" {
                    if let index = itemModels[currItemModelIndex].titleStickers.firstIndex(of: title) {
                        itemModels[currItemModelIndex].titleStickers.remove(at: index)
                    }
                }
            }
        }
        subtitleCount = itemModels[currItemModelIndex].titleStickers.count
        BFLog(2, message: "删除\(voiceModel?.wavFilePath ?? "")对应的字幕  后 count\(subtitleCount)")
        /// 重置进度
        ///
        DispatchQueue.main.async {[weak self] in
            guard let wself = self else { return }
            wself.recordBtn.isHidden = false

            if voiceModel?.currIndex == wself.currItemModelIndex {
                if wself.currMediaType == .Image {
                    wself.rscurrentManager.currentAssetProgress = CMTime(seconds: voiceModel?.startCMTime.seconds ?? 0, preferredTimescale: 1000)
                    if wself.currentAssetProgress.seconds >= wself.itemModels[wself.currItemModelIndex].materialDuraion.seconds {
                        wself.rscurrentManager.currentAssetProgress = CMTime(seconds: wself.itemModels[wself.currItemModelIndex].materialDuraion.seconds, preferredTimescale: 1000)
                    }
                    wself.resetCurrentProgress()
                }
            // 移除
                
                wself.resetAllIndirectionView()
            }
        }
        recorderManager?.voiceModel = nil
    }

    // 撤销
    @objc func withdrawAction() {
        pause()
        if let action = events.last {
            let jumpTime = action.timestamp
            if action.type == 2 {
                // 撤销录制
                if let modelIndex = itemModels[currItemModelIndex].voiceStickers.firstIndex(where: { mod in
                    CMTimeCompare(mod.startCMTime, action.timestamp) == 0
                }) {
                    // 移除音频
                    let model = itemModels[currItemModelIndex].voiceStickers[modelIndex]
                    itemModels[currItemModelIndex].voiceStickers.remove(at: modelIndex)
                    indirectionView?.deleteItem(index: modelIndex)

                    // 删除对应字幕
                    itemModels[currItemModelIndex].titleStickers.removeAll { m in
                        m.audioFilePath == model.wavFilePath
                    }

                    // 恢复被覆盖的音频
                    if let voiceTuples = action.deletedVoices{
                        itemModels[currItemModelIndex].voiceStickers.append(contentsOf: voiceTuples)
                        itemModels[currItemModelIndex].voiceStickers.sort { m1, m2 in
                            CMTimeCompare(m1.startCMTime, m2.startCMTime) < 0
                        }
                    }
                    
                    // 恢复被覆盖的字幕
                    if let titleTuples = action.deletedTittles{
                        itemModels[currItemModelIndex].titleStickers.append(contentsOf: titleTuples)
                        itemModels[currItemModelIndex].titleStickers.sort { m1, m2 in
                            CMTimeCompare(m1.timelineOut, m2.timelineOut) < 0
                        }
                    }
//                    jumpTime = model.startCMTime.seconds

                    if currMediaType == .Image {
                        itemModels[currItemModelIndex].materialDuraion = model.startCMTime
                    }else if currMediaType == .Camera {
                        let video = rscmanager.recordItem?.videoStickers.last
                        rscmanager.revertLast(modelIndex)
                        itemModels[currItemModelIndex].materialDuraion = video?.timelineCMIn ?? .zero
                    }
                }
            } else if action.type == 3 {
                // 恢复删除录音
                if var tuples = action.deletedVoices{
                    
                    if currMediaType != .Video {

                        tuples.sort { m1, m2 in
                            CMTimeCompare(m1.startCMTime, m2.startCMTime) < 0
                        }
                        
                        var firstTime = 0.0
                        var currDuration : CMTime = .zero
                        
                        if currMediaType == .Camera, let first = action.deletedCameras?.first, let end = action.deletedCameras?.last {
                            firstTime = first.timelineCMIn.seconds
                            currDuration = (end.timelineCMOut - first.timelineCMIn)
                        }else if currMediaType == .Image, let first = tuples.first, let end = tuples.last {
                            firstTime = first.startCMTime.seconds
                            currDuration = end.endCMTime - first.startCMTime
                        }
                        
                        BFLog(1, message: "恢复删除时长:\(currDuration.seconds)")
                        
                        itemModels[currItemModelIndex].materialDuraion = itemModels[currItemModelIndex].materialDuraion + currDuration
                        
                        // 恢复录音的时间点
                        let list = itemModels[currItemModelIndex].voiceStickers.filter({ mod in
                            CMTimeCompare(mod.startCMTime, CMTime(seconds: firstTime, preferredTimescale: 1000)) >= 0
                        })
                    
                        list.forEach { mod in
                            mod.startCMTime = mod.startCMTime + currDuration
                            mod.endCMTime = mod.endCMTime + currDuration
                            
                            // 恢复字幕的时间点
                            let titlsList = itemModels[currItemModelIndex].titleStickers.filter { tm in
                                tm.recordId == mod.recordId
                            }
                            
                            for titleM in titlsList {
                                titleM.timelineIn = titleM.timelineIn + currDuration
                                titleM.timelineOut = titleM.timelineOut + currDuration
                            }
                        }
                        
                        if currMediaType == .Camera {
                            rscmanager.cameraProgressV?.isHidden = false
                            // 恢复录像的时间点
                            _ = itemModels[currItemModelIndex].videoStickers.map { mod in
                                if CMTimeCompare(mod.timelineCMIn, CMTime(seconds: firstTime, preferredTimescale: 1000)) >= 0{
                                    mod.timelineCMIn = mod.timelineCMIn + currDuration
                                    mod.timelineCMOut = mod.timelineCMOut + currDuration
                                }
                            }
                            
                            // 摄像头录制的
                            if let videoTuple = action.deletedCameras {
                                itemModels[currItemModelIndex].videoStickers.append(contentsOf: videoTuple)
                                itemModels[currItemModelIndex].videoStickers.sort { m1, m2 in
                                    CMTimeCompare(m1.timelineCMIn, m2.timelineCMIn) < 0
                                }
                                itemModels[currItemModelIndex].thumbImgs.removeAll()
                                itemModels[currItemModelIndex].videoStickers.forEach { mod in
                                    if let thumImag = mod.thumImgs{
                                        itemModels[currItemModelIndex].thumbImgs.append(contentsOf: thumImag)
                                    }
                                }
                                DispatchQueue.main.async {[weak self] in
                                    guard let wself = self else { return }
                                    wself.rscmanager.cameraProgressV?.resetThumIV()
                                }
                            }
                        }
                    
                    }
                    
                    itemModels[currItemModelIndex].voiceStickers.append(contentsOf: tuples)
                    itemModels[currItemModelIndex].voiceStickers.sort { m1, m2 in
                        CMTimeCompare(m1.startCMTime, m2.startCMTime) < 0
                    }
                }
                // 恢复字幕
                
                if let titleTuples = action.deletedTittles, titleTuples.count > 0 {
                    itemModels[currItemModelIndex].titleStickers.append(contentsOf: titleTuples)
                    itemModels[currItemModelIndex].titleStickers.sort { m1, m2 in
                        CMTimeCompare(m1.timelineOut, m2.timelineOut) < 0
                    }
                }
                if currMediaType == .Image {
                    itemModels[currItemModelIndex].materialDuraion = itemModels[currItemModelIndex].voiceStickers.last?.endCMTime ?? .zero
                }

            } else {}
            events.removeLast()

            let dur = itemModels[currItemModelIndex].materialDuraion.seconds
            if dur > 0 {
                if currMediaType == .Image {
                    changeProgress(isBack: true, progress: Float(jumpTime.seconds))
                } else {
                    changeProgress(changCMTime: jumpTime)
                }
            } else {
                changeProgress(progress: Float(0))
            }
            isDragingProgressSlder = false
            currentPlayRecordIndex = -1
            BFLog(3, message: "重置播放index-\(#function) = \(currentPlayRecordIndex)")
            hadPrepareToPlayRecord = false
            progressThumV.progress = jumpTime.seconds

            if let event = events.last {
                changeWithDrawBtnLayout(event.type)
            } else {
                changeWithDrawBtnLayout(0)
                withDrawBtn.isEnabled = false
            }
            searchStopAtRecordRange()
            let itemModel = itemModels[currItemModelIndex]
            /// 重绘录音进度视图
            resetAllIndirectionView()
            // 如果是图片需重置播放按钮
            if itemModel.mediaType != .Video {
                playBtn.isSelected = itemModels[currItemModelIndex].voiceStickers.count <= 0
                playBtn.isHidden = playBtn.isSelected
            }
        }
    }

    @objc func changeVoiceAction() {
        //        nextActionHandle?()
        pause()
    }
    
    @objc func playbtnDown(btn:UIButton){
        let vv = UIView(frame: CGRect(x: 0, y: 0, width: 0.5, height: 0.5))
        vv.backgroundColor = UIColor.init(white: 1, alpha: 0.3)
        btn.setImage(vv.graphicsGetImage(), for: .normal)
    }
    
    @objc func playbtnCancle(btn:UIButton){
        btn.setImage(imageInRecordScreenKit(by: "preview_play"), for: .normal)
    }

    @objc public func playVideo(btn: UIButton) {
        btn.setImage(imageInRecordScreenKit(by: "preview_play"), for: .normal)

        if currMediaType != .Video && itemModels[currItemModelIndex].voiceStickers.count <= 0 {
            BFLog(message: "图片或摄像头没有录音无法播放")
            btn.isSelected = true
            return
        }
        
        btn.isSelected = !btn.isSelected
        if !btn.isSelected {
            pause()
            searchStopAtRecordRange()
        } else {
            events.append(WithDrawModel(type: 1, timestamp: currentAssetProgress))
            changeWithDrawBtnLayout(0)
            play()
        }
    }
    
    // MARK: - 进度条拖动回调
    
    func thumbViewEnded(progress: Float) {
        changeProgress(progress: progress)
        isDragingProgressSlder = false
        currentPlayRecordIndex = -1
        hadPrepareToPlayRecord = false
        BFLog(1, message: "isDragingProgressSlder : \(isDragingProgressSlder)")
        BFLog(3, message: "重置播放index-\(#function) = \(currentPlayRecordIndex)")
        
        isEndPlay = (progress == 1)
        searchStopAtRecordRange()
        
        changeWithDrawBtnLayout(0)
        recorderManager?.voiceModel = nil
        
        rscurrentManager.currentAssetProgress = currentAssetProgress
        rscurrentManager.dragEnd()
    }
    
    func thumViewCallBack(_ vv: BFProgressBaseView) {
        vv.dragStartHandle = { [weak self] in
            guard let wself = self else { return }
            
            wself.isDragingProgressSlder = true
            wself.pause()
        }
        vv.dragScrollProgressHandle = { [weak self] isStart, process in
            DispatchQueue.main.async { [weak self] in
                guard let wself = self else { return }
                
                if isStart {
                    wself.events.append(WithDrawModel(type: 0, timestamp: wself.currentAssetProgress))
                }

                wself.isDragingProgressSlder = true

                wself.changeProgress(progress: process)
                
                if wself.currMediaType != .Video && CMTimeCompare(wself.rscurrentManager.recordItem?.materialDuraion ?? .zero, .zero) != 0 {
                    wself.recordBtn.isHidden = true
                    wself.deleteRecordBtn.isHidden = false
                }
            }
        }
        vv.dragEndHandle = { [weak self] process in
            BFLog(1, message: "stop drag")
            DispatchQueue.main.async { [weak self] in
                guard let wself = self else { return }

                wself.thumbViewEnded(progress: process)
            }
        }
    }

    // MARK: - 逻辑处理
    // 是否吸附在录音首尾处
    func searchStopAtRecordRange(needAdsorb: Bool = false) {
        // TODO: 滑动，播放暂停，撤销时，判断是否停止录音区间，是则删除相关录音，画笔，头像，字幕
        
        var startTime: CMTime?
        var endTime: CMTime?
        isStopAtRecordRange = -1
        
        isEndPlay = (CMTimeCompare(currentAssetProgress, rscurrentManager.recordItem?.materialDuraion ?? .zero) >= 0)

        if currMediaType == .Camera {
            let elems = itemModels[currItemModelIndex].videoStickers.enumerated().filter { elem in
                CMTimeCompare(elem.1.timelineCMIn, currentAssetProgress) <= 0 && elem.1.timelineCMOut > currentAssetProgress
            }
            if elems.count > 0 {
                startTime = elems[0].1.timelineCMIn
                endTime = elems[0].1.timelineCMOut
                isStopAtRecordRange = elems[0].0
            }
        }else {
            let elems = itemModels[currItemModelIndex].voiceStickers.enumerated().filter { elem in
                CMTimeCompare(elem.1.startCMTime, currentAssetProgress) <= 0 && elem.1.endCMTime > currentAssetProgress
            }
            if elems.count > 0 {
                startTime = elems[0].1.startCMTime
                endTime = elems[0].1.endCMTime
                isStopAtRecordRange = elems[0].0
            }
        }

        if startTime != nil && endTime != nil {
            //  TODO: 停在了录音区间,显示删除按钮
            if needAdsorb {
                if fabs((endTime! - currentAssetProgress).seconds) < 0.5 {
                    BFLog(1, message: "吸附在录音结尾, \(endTime!.seconds)")
                    //                changeWithDrawBtnLayout(false)
                    changeProgress(changCMTime: endTime!)
                    progressThumV.progress = endTime!.seconds

                    searchStopAtRecordRange(needAdsorb: false)
                    return
                } else {
                    if fabs((startTime! - currentAssetProgress).seconds) < 0.5 {
                        BFLog(1, message: "吸附在录音开始")
                        //                    changeWithDrawBtnLayout(true)
                        changeProgress(changCMTime: startTime!)
                        progressThumV.progress = startTime!.seconds
                    }
                    deleteRecordBtn.isHidden = false
                    recordBtn.isHidden = true
                    // add by ak 这里 isHidden 无效

                    BFLog(1, message: "停在了录音区间里 \(isStopAtRecordRange), currTime:\(currentAssetProgress.seconds), 录音范围：\(startTime!.seconds) - \(endTime!.seconds)")
                }
            } else {
                updateRecordBtnStatus(true)

                BFLog(1, message: "停在了录音区间里 \(isStopAtRecordRange), currTime:\(currentAssetProgress.seconds), 录音范围：\(startTime!.seconds) - \(endTime!.seconds)")
            }

        } else {
            updateRecordBtnStatus(false)

            
            isStopAtRecordRange = -1
            BFLog(1, message: "停在了录音区间外 \(isStopAtRecordRange), currTime:\(currentAssetProgress.seconds)")

        }
        
        updatePlayBtnStatus()
    }
    func updatePlayBtnStatus(){
        switch currMediaType {
        case .Image:
            playBtn.isSelected = ((rsimanager.recordItem?.voiceStickers.count ?? 0) <= 0)
            playBtn.isHidden = playBtn.isSelected
            
        case .Video:
            playBtn.isSelected = false
            playBtn.isHidden = false
        case .Camera:
            playBtn.isSelected = isEndPlay ? true : ((rscmanager.recordItem?.voiceStickers.count ?? 0) <= 0)
            playBtn.isHidden = playBtn.isSelected
        }
    }
    
    func updateRecordBtnStatus(_ show:Bool? = nil){
        let su = show ?? false
        switch currMediaType {
        case .Image:
            recordBtn.isHidden = isEndPlay ? false : su
        case .Video:
            recordBtn.isHidden = isEndPlay ? true : su
        case .Camera:
            recordBtn.isHidden = isEndPlay ? false : su
        }
        
        deleteRecordBtn.isHidden = isEndPlay ? true : !su
    }

    // MARK: - 权限申请

    func checkStatus(show: Bool = true) -> Bool {
        let status = AVCaptureDevice.authorizationStatus(for: .audio)
        switch status {
        case .denied, .restricted:

            BFLog(message: "麦克风权限被拒绝，请在设置中打开")
            if show {
                let remindData = BFBaseModel()
                remindData.title = "grant_micpass".BFLocale
                remindData.summary = ""
                let remindView = BFRemindView(frame: CGRect(x: 0, y: 0, width: cScreenWidth, height: cScreenHeigth))
                remindView.isBanned = true
                remindView.confirmBtn.setTitle("grant_open".BFLocale, for: .normal)
                remindView.confirmBtn.setTitleColor(UIColor.hexColor(hexadecimal: "#389AFF"), for: .normal)
                UIApplication.shared.keyWindow?.addSubview(remindView)
                remindView.remindData = remindData
                remindView.cancelBtn.setTitleColor(UIColor.hexColor(hexadecimal: "#389AFF"), for: .normal)
                remindView.remindBlock = { item, _ in
                    if item.tag == 2 {
                        openAppSetting()
                    }
                }
            }
            return false
        case .authorized:
            return true
        case .notDetermined:
            if show {
                requestMicroPhoneAuth()
            }
            return false
        default:
            break
        }

        return false
    }

    func requestMicroPhoneAuth() {
        AVCaptureDevice.requestAccess(for: .audio) { granted in
            if !granted {
                BFLog(message: "麦克风权限被拒绝，请在设置中打开")
            }
        }
    }

    // MARK: - 音视频处理

    func playRecord(at currentT: CMTime, periodicTimeObserver: @escaping (_ time: CMTime, _ currentItem: AVPlayerItem) -> Void, didPlayToEndTime: @escaping (_ recordedInfo: (Int, PQVoiceModel)?, _ currentItem: AVPlayerItem?) -> Void, playFailed _: @escaping (_ recordedInfo: (Int, PQVoiceModel)?, _ currentItem: AVPlayerItem?) -> Void) {
//        if currentPlayRecordIndex >= 0 {
//            if assetPlayer.volume != haveSpeakVolume{
//                assetPlayer.volume = haveSpeakVolume
//            }
//        }else{
//            if assetPlayer.volume != noSpeakVolume {
//                assetPlayer.volume = noSpeakVolume
//            }
//        }

        if itemModels[currItemModelIndex].voiceStickers.first(where: { m in
            CMTimeCompare(m.startCMTime, currentT) <= 0 && CMTimeCompare(currentT, m.endCMTime) <= 0
        }) != nil {
            if assetPlayer.volume != haveSpeakVolume {
                assetPlayer.volume = haveSpeakVolume
            }
        } else {
            if assetPlayer.volume != noSpeakVolume {
                assetPlayer.volume = noSpeakVolume
            }
        }
//        BFLog(1, message: "volume:\(assetPlayer.volume ?? -1)")

        if currentPlayRecordIndex == -3 { // 刚录音完，不需要播放
            return
        }

        // 先排序，再查找下一个需要播放的录音
        let list = itemModels[currItemModelIndex].voiceStickers.sorted { m1, m2 in
            CMTimeCompare(m1.startCMTime, m2.startCMTime) < 0
        }
        let (shouldPlayRecordIndex, recordedAudio) = list.enumerated().first { model in
            CMTimeCompare(model.1.endCMTime, currentT) > 0
        } ?? (-1, nil)

        // 没找到，代表后边没有录音需求了
        guard let recordedAudio = recordedAudio, recordedAudio.wavFilePath.count > 0 else {
            BFLog(3, message: "未找到可播放录音")
            return
        }
//        BFLog(1, message: "当前时间：\(CMTimeGetSeconds(currentT)), 找到的音频：\(recordedAudio.startCMTime.seconds) ～ \(recordedAudio.endCMTime.seconds), \(recordedAudio.wavFilePath ?? "")")

        // 创建播放器
        if (recordPlayer.currentItem?.asset as? AVURLAsset)?.url.lastPathComponent != (recordedAudio.wavFilePath as NSString?)?.lastPathComponent {
            let newItem = AVPlayerItem(url: URL(fileURLWithPath: recordedAudio.wavFilePath))
            BFLog(1, message: "录音播放器初始化：\("replace item")")

            recordPlayer.pause()
            if let playItem = recordPlayer.currentItem {
                NotificationCenter.default.removeObserver(self, name: .AVPlayerItemDidPlayToEndTime, object: playItem)
            }
            recordPlayer.replaceCurrentItem(with: newItem)
            
            recordPlayer.volume = 1
            currentPlayRecordIndex = -1
            hadPrepareToPlayRecord = false
            BFLog(1, message: "录音播放器初始化(有时候不准)")
            BFLog(3, message: "重置播放index-\(#function) = \(currentPlayRecordIndex)")
            NotificationCenter.default.addObserver(forName: .AVPlayerItemDidPlayToEndTime, object: newItem, queue: .main) { [weak self] _ in
                guard let wself = self else {
                    BFLog(3, message: "wself为空AVPlayerItemDidPlayToEndTime")
                    return
                }
                wself.hadPrepareToPlayRecord = false
                wself.currentPlayRecordIndex = -1
//                wself.recordPlayer.volume = 0
//                wself.assetPlayer.volume = wself.noSpeakVolume
                BFLog(3, message: "播放结束")
                didPlayToEndTime((shouldPlayRecordIndex, recordedAudio), newItem)
            }
//            recordPlayer.removeTimeObserver(recordPlayerTimeObserver as Any)
//            recordPlayerTimeObserver?.invalidate()
            recordPlayerTimeObserver = recordPlayer.addPeriodicTimeObserver(forInterval: CMTime(value: 1, timescale: 1000), queue: DispatchQueue.global()) { [weak self, weak recordPlayer] time in
                guard let wself = self, let rPlay = recordPlayer else {
                    BFLog(3, message: "wself为空")
                    return
                }
                BFLog(1, message: "当前播放--- \(time.seconds), \(rPlay.currentItem?.duration.seconds ?? 0)")
                if CMTimeGetSeconds(wself.currenStartPlayTime) <= 0 {
                    BFLog(message: "重新更新开始播放进度\(#function)-\(wself.currenStartPlayTime.seconds)")
                    wself.currenStartPlayTime = time
                }
                let progressTime = CMTime(seconds: CMTimeGetSeconds(time) - CMTimeGetSeconds(wself.currenStartPlayTime), preferredTimescale: 1000)
                BFLog(message: "progressTime进度\(#function)-\(progressTime.seconds)")
                periodicTimeObserver(progressTime, newItem)
//                if (rPlay.currentItem?.currentTime().seconds ?? 0) > (rPlay.currentItem?.duration.seconds ?? 0) - 0.1 {
//                    didPlayToEndTime((shouldPlayRecordIndex, recordedAudio), newItem)
//                }
            } as? NSKeyValueObservation
        }
        if currMediaType == .Video {
            videoMaterialRecordPlay(at: currentT, shouldPlayRecordIndex: shouldPlayRecordIndex, recordedAudio: recordedAudio)
        } else {
            imageMaterialRecordPlay(at: currentT, shouldPlayRecordIndex: shouldPlayRecordIndex, recordedAudio: recordedAudio)
        }
        BFLog(1, message: "应当播放：\(shouldPlayRecordIndex), 当前播放：\(currentPlayRecordIndex)")
    }

    /// 视频素材录音的播放
    /// - Parameters:
    ///   - currentT: <#currentT description#>
    ///   - shouldPlayRecordIndex: <#shouldPlayRecordIndex description#>
    ///   - recordedAudio: <#recordedAudio description#>
    func videoMaterialRecordPlay(at currentT: CMTime, shouldPlayRecordIndex: Int, recordedAudio: PQVoiceModel) {
        //        if recordPlayer.currentItem?.duration.timescale == 0 {
        //            BFLog(1, message: "时间timescale  == 0")
        //            playFailed((shouldPlayRecordIndex, recordedAudio) as? (Int, PQVoiceModel), recordPlayer.currentItem)
        //            return
        //        }
        synced(currentPlayRecordIndex) { [weak self] in
            guard let wself = self else {
                return
            }
//            BFLog(1, message: "判断是否开始录音播放** hadPrepareToPlayRecord:\(hadPrepareToPlayRecord), currentPlayRecordIndex:\(currentPlayRecordIndex), isNormalPlaying :\(wself.isNormalPlaying),\(recordPlayer.currentItem?.duration.timescale ?? 0),\(CMTimeGetSeconds(currentT) >= recordedAudio.startCMTime.seconds),\(CMTimeGetSeconds(currentT) <= recordedAudio.endCMTime.seconds - 0.2)")

            if !wself.hadPrepareToPlayRecord,
               CMTimeGetSeconds(currentT) >= (recordedAudio.startCMTime.seconds - 0.1),
               CMTimeGetSeconds(currentT) <= recordedAudio.endCMTime.seconds - 0.2 // 这个条件是避免录音结束后有小幅度回退导致播放最新录音
            {
                if wself.currMediaType == .Video, wself.recordPlayer.currentItem?.duration.timescale == 0 {
                    return
                }
                // 应当开始播放了
                // 两个逻辑：如果在播，则跳过;如果暂停拖动到中间，则seek
                if wself.currentPlayRecordIndex == -1, wself.isNormalPlaying {
                    let second = currentT - recordedAudio.startCMTime
                    DispatchQueue.main.async { [weak self] in
                        
                        guard let wself = self else { return }
                        
                        wself.recordPlayer.seek(to: second, toleranceBefore: CMTime(value: 1, timescale: 1_000_000), toleranceAfter: CMTime(value: 1, timescale: 1_000_000), completionHandler: { [weak self] finished in
                            
                            guard let wself = self else { return }
                            
                            if finished, wself.isNormalPlaying {
                                wself.recordPlayer.play()
//                                wself.recordPlayer.volume = 1
//                                wself.assetPlayer.volume = wself.haveSpeakVolume ?? 0
                                BFLog(1, message: "录音开始播放 playing, \(second), \(CMTimeGetSeconds(wself.recordPlayer.currentItem?.duration ?? .zero)), \(wself.recordPlayer.currentItem?.currentTime().seconds ?? 0)")
                            }
                        })
                    }
                    currentPlayRecordIndex = shouldPlayRecordIndex
                    hadPrepareToPlayRecord = true
                    BFLog(3, message: "重置播放index-\(#function) = \(currentPlayRecordIndex)")
                    BFLog(1, message: "录音开始播放2, \(second), \(CMTimeGetSeconds(recordPlayer.currentItem?.duration ?? .zero)),index = \(currentPlayRecordIndex)")
                }
            }
        }
    }

    /// 图片素材录音的播放
    /// - Parameters:
    ///   - currentT: <#currentT description#>
    ///   - shouldPlayRecordIndex: <#shouldPlayRecordIndex description#>
    ///   - recordedAudio: <#recordedAudio description#>
    func imageMaterialRecordPlay(at currentT: CMTime, shouldPlayRecordIndex _: Int, recordedAudio: PQVoiceModel) {
        synced(currentPlayRecordIndex) { [weak self] in
            guard let wself = self, wself.isNormalPlaying else {
                return
            }

            if CMTimeGetSeconds(currentT) >= (recordedAudio.startCMTime.seconds - 0.1),
               CMTimeCompare(currentT, recordedAudio.endCMTime) <= 0 {
                // 应当开始播放了
                // 两个逻辑：如果在播，则跳过;如果暂停拖动到中间，则seek
                if wself.isNormalPlaying {
                    let second = currentT - recordedAudio.startCMTime
                    wself.recordPlayer.seek(to: second, toleranceBefore: CMTime(seconds: 1, preferredTimescale: 1000), toleranceAfter: CMTime(seconds: 1, preferredTimescale: 1000), completionHandler: { [weak self] isFinished in
                        guard let wself = self else { return }
                        
                        wself.recordPlayer.play()
                    })
                    BFLog(3, message: "录音开始播放2, \(second), \(CMTimeGetSeconds(recordPlayer.currentItem?.duration ?? .zero)),index = \(wself.currentPlayRecordIndex)")
                }
            }
        }
    }

    func play() {
        BFLog(1, message: "开始播放 \(currentAssetProgress.seconds)")
        recorderManager?.voiceModel = nil

        itemModels[currItemModelIndex].titleStickers.sort { m1, m2 in
            CMTimeCompare(m1.timelineIn, m2.timelineIn) < 0
        }

        isNormalPlaying = true
        if isEndPlay || (currMediaType == .Image && CMTimeCompare(currentAssetProgress, itemModels[currItemModelIndex].materialDuraion) >= 0) {
            isEndPlay = false
            assetPlayer.seek(to: CMTime.zero)
            progressThumV.progress = 0
            currentPlayRecordIndex = -1
            BFLog(3, message: "重置播放index-\(#function) = \(currentPlayRecordIndex)")

            rscurrentManager.currentAssetProgress = .zero
        }
        
        rscurrentManager.play()
        
        deleteRecordBtn.isHidden = true

        subtitleBtn.isHidden = true
        soundSettingBtn.isHidden = true
        withDrawBtn.isHidden = true
        recordBtn.isHidden = true
    }

    public func pause() {
//        BFLog(1, message: "暂停播放")
        isNormalPlaying = false

        subtitleBtn.isHidden = false
        soundSettingBtn.isHidden = false
        withDrawBtn.isHidden = false
        
        assetPlayer.pause()
        recordPlayer.pause()
        recordStartPlayTime = CMTime.zero
        pauseTime = currentAssetProgress.seconds
        currentPlayRecordIndex = -1
        hadPrepareToPlayRecord = false

        // 暂停状态
        playBtn.isSelected = (currMediaType == .Image && rscurrentManager.recordItem?.voiceStickers.count ?? -1 <= 0) || (currMediaType == .Camera && isEndPlay)
        playBtn.isHidden = playBtn.isSelected
        rscurrentManager.pause()
    }

    public func fetchMaterial(_ phAsset:[PHAsset]) {
        
        var lastCount = itemModels.count

        if lastCount > 0
            && itemModels[lastCount-1].mediaType == .Camera
            && itemModels[lastCount-1].voiceStickers.count == 0{
            removeCameroCell()
            lastCount -= 1
        }
        
        var indexPaths = [IndexPath]()
        
        if phAsset.count > 0 {
            for (index, asset) in phAsset.enumerated() {
                let itemModel = BFRecordItemModel()
                itemModel.index = index
                if asset.title == "record.camera"{
                    itemModel.mediaType = .Camera
                }else{
                    itemModel.initOriginData(phasset: asset)
                    if asset.mediaType != .video {
                        itemModel.coverPath = asset.localPath
                    }
                    itemModel.fetchCoverImgCallBack = { [weak self, weak itemModel] img in
                        guard let wself = self, let item = itemModel else {
                            return
                        }
                        if let cell = wself.collectionView.cellForItem(at: IndexPath(item: wself.currItemModelIndex, section: 0)) as? BFImageCoverViewCell, cell.recordItem?.coverImg == img {
                            cell.addData()
                            if item.mediaType != .Camera {
                                wself.progressThumV.recordItem?.coverImg = item.coverImg
                                wself.progressThumV.isHidden = false
                            }
                        }
                    }
                    itemModel.fetchPlayItemCallBack = { [weak self, weak itemModel] _ in
                        guard let wself = self, let item = itemModel else {
                            return
                        }
                        guard item.playItem != nil  else {
                            BFLog(message: "视频获取失败:\(item.index)")
                            cShowHUB(superView: nil, msg: "option_fail_obtain".BFLocale)
                            return
                        }
                        if let cell = wself.collectionView.cellForItem(at: IndexPath(item: wself.currItemModelIndex, section: 0)) as? BFImageCoverViewCell, cell.recordItem?.localPath == item.localPath {
                                wself.rscurrentManager.resetEnv()
                                if wself.currMediaType == .Video {
                                    (cell as? BFVideoCoverViewCell)?.rotationView()
                                }
                        }
                    }
                }
                indexPaths.append(IndexPath(row: itemModels.count, section: 0))
                itemModels.append(itemModel)
            }
        }
        collectionView.reloadData()
//        collectionView.contentOffset = CGPoint(x: cScreenWidth * CGFloat(lastCount) , y: 0.0)
        collectionView.scrollToItem(at: IndexPath(row: lastCount, section: 0), at: .left, animated: false)
        DispatchQueue.main.asyncAfter(deadline: .now() + 0.1) {[weak self] in
            guard let wself = self else { return }
            
            wself.endScrollItem(page: lastCount)
        }
        
//        // 暂停状态--如果是图片素材同时没有录音文件时不显示播放按钮
//        playBtn.isSelected = (itemModels.first?.mediaType != .Video && (itemModels.first?.voiceStickers.count ?? 0) <= 0)
//        playBtn.isHidden = playBtn.isSelected
    }

    func setCoverImage(img: UIImage) {
        if let pic = GPUImagePicture(image: img) {
            let filter = GPUImageFilter()
            pic.addTarget(filter)
            filter.addTarget(playView)
            pic.processImage()
        }
    }

    func setAudioPlay(item: AVPlayerItem?) {
        BFLog(message: "设置播放器item：\(String(describing: item))")
        guard let item = item else {
            return
        }
        if let playItem = assetPlayer.currentItem {
            NotificationCenter.default.removeObserver(self, name: .AVPlayerItemDidPlayToEndTime, object: playItem)
            assetPlayer.replaceCurrentItem(with: item)
            BFLog(message: "设置播放器playItem-替换：\(String(describing: item))")
        } else {
            assetPlayer.replaceCurrentItem(with: item)
            BFLog(message: "设置播放器item-替换：\(String(describing: item))")
            if avplayerTimeObserver != nil{
                avplayerTimeObserver?.invalidate()
            }
            avplayerTimeObserver = assetPlayer.addPeriodicTimeObserver(forInterval: CMTime(value: 1, timescale: 1000), queue: DispatchQueue.global()) { [weak self] time in
                // 进度监控
                guard let wself = self else { return }
                wself.periodicTimeObserver(item: item, time: time)
                if wself.isNormalPlaying {
                    // 播放录音
                    wself.startPlayRecord(time: time)
                }
            } as? NSKeyValueObservation
        }

        NotificationCenter.default.addObserver(forName: .AVPlayerItemDidPlayToEndTime, object: assetPlayer.currentItem, queue: .main) { [weak self] _ in
            guard let wself = self else {
                return
            }
            
            // 非录音的播放状态，结束时自动跳下一段落
            if !wself.isRecording, wself.currItemModelIndex < (wself.itemModels.count - 1) {
            } else {} //

            wself.isNormalPlaying = false
            wself.isEndPlay = true
            wself.updatePlayBtnStatus()


            if wself.isRecording {
                wself.endRecord()
                cShowHUB(superView: nil, msg: "tips_full".BFLocale)
            }
        }
        
    }

    /// 音视频播放进度-视频跟图片单独处理
    /// - Parameter CMTime: <#CMTime description#>
    func periodicTimeObserver(item: AVPlayerItem?, time: CMTime) {
        // 进度监控
        //        if time.seconds < pauseTime {
        //            return
        //        }
        if currMediaType == .Video {
            if CMTimeGetSeconds(item?.duration ?? CMTime.zero) > 0 {
                rscurrentManager.currentAssetProgress = CMTime(seconds: time.seconds, preferredTimescale: 1000)
//                BFLog(1, message: "video curr:\(CMTimeGetSeconds(currentAssetProgress))")
                DispatchQueue.main.async { [weak self] in
                    guard let wself = self else { return }
                    
                    BFLog(message: "更新录音进度\(#function)-\(wself.currentAssetProgress.seconds )")
                    wself.progreddL.text = String(format: "%@", CMTimeGetSeconds(time).formatDurationToHMS())
                    let su = !wself.isDragingProgressSlder || wself.isRecording || wself.isNormalPlaying
                    if su { // 不拖动，正常播放和录音时更新进度条
                        wself.progressThumV.progress = time.seconds
                    }
                    // 更新字幕
                    if !wself.isRecording {
                        wself.updateSubtitle(time: time)
                    }
                }
            }
        }

        if isRecording {
            let startTime = recorderManager?.voiceModel?.startCMTime.seconds ?? 0
            let progress = (currentAssetProgress - (recorderManager?.voiceModel?.startCMTime ?? .zero)).seconds // - ratioX
            // 使用播放器的进度来画线，因为进度是跟着播放器来了
            DispatchQueue.main.async { [weak self] in
                guard let wself = self else { return }
                
                wself.indirectionView?.setProgress(start: startTime, progress: max(0, progress))
            }
        }
    }

    // MARK: - 录音对应图像绘制
    

    // 撤销按钮修改title，重绘
    func changeWithDrawBtnLayout(_ type: Int) {
        switch type {
        case 3:
            withDrawBtn.setTitle("record_recover".BFLocale, for: .normal)
        case 2:
            withDrawBtn.setTitle("record_undo".BFLocale, for: .normal)
        default:
            withDrawBtn.setTitle("record_back".BFLocale, for: .normal)
        }
        withDrawBtn.imageEdgeInsets = UIEdgeInsets(top: -withDrawBtn.imageView!.height, left: 0, bottom: 0, right: -withDrawBtn.titleLabel!.width)
        withDrawBtn.titleEdgeInsets = UIEdgeInsets(top: withDrawBtn.titleLabel!.height + 2, left: -withDrawBtn.imageView!.width, bottom: 0, right: 0)

        withDrawBtn.setNeedsLayout()
        withDrawBtn.layoutIfNeeded()
    }

    // 通过缩略图进度条控制播放进度
    // progress : 图片且isback为true时，表示是时长；  其他情况： 0 - 1，百分比
    func changeProgress(isBack: Bool = false, progress: Float = -1, changCMTime: CMTime = .zero) {
        var newProgress = progress
        if progress.isNaN || progress.isInfinite {
            newProgress = 0
        }
        if currMediaType == .Video {
            let duration = itemModels[currItemModelIndex].materialDuraion
            if duration.seconds > 0 {
                if progress == -1 {
                    rscurrentManager.currentAssetProgress = changCMTime
                } else {
                    rscurrentManager.currentAssetProgress = CMTime(seconds: Double(newProgress) * duration.seconds, preferredTimescale: 1000)
                }
                DispatchQueue.main.async { [weak self] in
                    guard let wself = self else { return }
                    
                    BFLog(message: "更新录音进度\(#function)- \(wself.currentAssetProgress.seconds)")
                    wself.progreddL.text = String(format: "%@", CMTimeGetSeconds(wself.currentAssetProgress).formatDurationToHMS())
                }
                assetPlayer.seek(to: currentAssetProgress, toleranceBefore: CMTime(value: 1, timescale: 1_000_000), toleranceAfter: CMTime(value: 1, timescale: 1_000_000)) { _ in
                }
            }
            
            let list = rscurrentManager.recordItem?.voiceStickers.filter({ model in
                CMTimeCompare(model.startCMTime, currentAssetProgress) <= 0 && CMTimeCompare(currentAssetProgress, model.endCMTime) < 0
            })
            if (list?.count ?? 0) > 0 {
                recordBtn.isHidden = true
                deleteRecordBtn.isHidden = false
            }else{
                recordBtn.isHidden = false
                deleteRecordBtn.isHidden = true
            }
            
        } else {
            rscurrentManager.currentAssetProgress = isBack ? CMTime(seconds: Double(newProgress), preferredTimescale: 1000) : ((progress != -1) ? CMTime(seconds: (Double(newProgress) * itemModels[currItemModelIndex].materialDuraion.seconds), preferredTimescale: 1000) : changCMTime)
            if currentAssetProgress.seconds > itemModels[currItemModelIndex].materialDuraion.seconds {
                rscurrentManager.currentAssetProgress = itemModels[currItemModelIndex].materialDuraion
            }
            DispatchQueue.main.async { [weak self] in
                guard let wself = self else { return }
                
//                BFLog(1,message: "更新进度- \(wself.currentAssetProgress.seconds)")
                wself.progreddL.text = String(format: "%@", CMTimeGetSeconds(self!.currentAssetProgress).formatDurationToHMS())
            }
            BFLog(message: "progress = \(progress),currentAssetProgress = \(currentAssetProgress.seconds),materialDuraion = \(itemModels[currItemModelIndex].materialDuraion.seconds)")

            // add by ak 图片素材拖动指针时同时更新字幕显现数据
            updateSubtitle(time: currentAssetProgress)
        }
        
    }

    // 画进度线
    func drawProgressIndication(progress: Double) {
        if currMediaType != .Camera, indirectionView == nil {
            var percenWidth: CGFloat = 0
            if currMediaType == .Image {
                percenWidth = progressThumV.thumbImageWidth / 2.0
            } else {
                percenWidth = progressThumV.progessIndicateBackV.frame.width / CGFloat(rscurrentManager.recordItem?.materialDuraion.seconds ?? 0)
            }
            // progressThumV.progressView.contentSize
            // progressThumV.progessIndicateBackV.origin
            indirectionView = BFIndirectionProgressView(frame: CGRect(x: progressThumV.progessIndicateBackV.x, y: 44, width: progressThumV.progressView.contentSize.width, height: 6), percenWidth: percenWidth, totalDuration: rscurrentManager.recordItem?.materialDuraion.seconds ?? 0)
            progressThumV.progressView.addSubview((indirectionView)!)
        }
        // 更新录制进度
        // 注：视频无法以录制进度驱动，因当录音开始录音时播放器还未播放，导致进度不一致
        // 注：在录制停止时，视频播放器进度依然在走，误差在80毫秒左右
        if isRecording, currMediaType == .Image {
            let startTime = recorderManager?.voiceModel?.startCMTime.seconds ?? 0
            // 使用播放器的进度来画线，因为进度是跟着播放器来了
            indirectionView?.setProgress(start: startTime, progress: progress)
        }
        if currMediaType == .Image {
            imageRecordProgress(isRecord: true, progress: progress)
        }
    }
}

extension BFRecordScreenController: GPUImageMovieDelegate {
    public func didCompletePlayingMovie() {
        BFLog(1, message: "播放结束")
        currentPlayRecordIndex = -1
        BFLog(3, message: "重置播放index-\(#function) = \(currentPlayRecordIndex)")
    }
}

extension BFRecordScreenController: AVAudioRecorderDelegate {
    public func audioRecorderDidFinishRecording(_: AVAudioRecorder, successfully _: Bool) {
        BFLog(1, message: "录音结束")
    }
}

extension BFRecordScreenController: AVAudioPlayerDelegate {
    public func audioPlayerDidFinishPlaying(_: AVAudioPlayer, successfully _: Bool) {
        BFLog(1, message: "录音播放结束")
    }
}

// MARK: - UICollectionViewDelegate

/// UICollectionViewDelegate
extension BFRecordScreenController: UICollectionViewDelegate, UICollectionViewDataSource, UIScrollViewDelegate {
    public func collectionView(_: UICollectionView, numberOfItemsInSection _: Int) -> Int {
        return itemModels.count
    }

    public func collectionView(_ collectionView: UICollectionView, cellForItemAt indexPath: IndexPath) -> UICollectionViewCell {
        let recordItem = itemModels[indexPath.item]
        var cell: BFImageCoverViewCell!

        switch recordItem.mediaType {
        case .Image:
            cell = BFImageCoverViewCell.gpuImageViewCell(collectionView: collectionView, indexPath: indexPath)
        case .Video:
            cell = BFVideoCoverViewCell.gpuVideoViewCell(collectionView: collectionView, indexPath: indexPath)
        case .Camera:
            cell = BFCameraCoverViewCell.gpuCamraViewCell(collectionView: collectionView, indexPath: indexPath)
        default:
            break
        }
        
        cell.btnClickHandle = { [weak self] _, _ in
            guard let wself = self else {
                return
            }
            wself.playVideo(btn: wself.playBtn)
        }
        cell.recordItem = recordItem
        return cell
    }

    public func collectionView(_: UICollectionView, didSelectItemAt _: IndexPath) {}

    public func collectionView(_ collectionView: UICollectionView, willDisplay cell: UICollectionViewCell, forItemAt indexPath: IndexPath) {
        if let currCell = cell as? BFVideoCoverViewCell {
            currCell.rotationView()
        }
 
    }
    
    public func collectionView(_ collectionView: UICollectionView, didEndDisplaying cell: UICollectionViewCell, forItemAt indexPath: IndexPath) {
        if let currCell = cell as? BFVideoCoverViewCell {
            currCell.rotationView()
        }
    }
    public func scrollViewWillBeginDragging(_: UIScrollView) {
        BFLog(1, message: "开始滚动")
        if currMediaType == .Camera && rscurrentManager.recordItem?.voiceStickers.count == 0{
            collectionView.isScrollEnabled = false
            showDeleteCamera()
            return
        }
        recordBtn.isEnabled = false
    }
    
    func showDeleteCamera(){
        let alertV = UIAlertController(title: nil, message: "shoot_tips_switch".BFLocale, preferredStyle: .alert)
        let cancelAction = UIAlertAction(title: "option_cancel".BFLocale, style: .default) {[weak self] _ in
            guard let wself = self else { return }
            wself.collectionView.isScrollEnabled = true
        }
        let okAction = UIAlertAction(title: "option_ok".BFLocale, style: .default) { [weak self] _ in
            guard let wself = self else { return }
            wself.collectionView.isScrollEnabled = true
            wself.removeCameroCell()
            wself.changeItemHandle?(-3)
        }
        alertV.addAction(cancelAction)
        alertV.addAction(okAction)
        
        self.present(alertV, animated: true, completion: nil)
    }

    public func scrollViewWillEndDragging(_ scrollView: UIScrollView, withVelocity velocity: CGPoint, targetContentOffset: UnsafeMutablePointer<CGPoint>) {
        let page = Int((targetContentOffset.pointee.x + scrollView.frame.width / 2) / scrollView.frame.width)
        
        if page != currItemModelIndex{
            endScrollItem(page: page)
        }

        BFLog(1, message: "将要停止拖动， page:\(currItemModelIndex) -> \(page)")
        
    }
    
    public func scrollViewDidEndDecelerating(_ scrollView: UIScrollView) {
        endScroll()
    }
    
    public func scrollViewDidEndDragging(_ scrollView: UIScrollView, willDecelerate decelerate: Bool) {
        if !decelerate {
            endScroll()
        }
    }
    
    func endScroll() {
        recordBtn.isEnabled = true
    }
    
    func removeCameroCell(){
        itemModels.remove(at: currItemModelIndex)
        UIView.performWithoutAnimation {[weak self] in
            guard let wself = self else { return }
            
            wself.collectionView.deleteItems(at: [IndexPath(row: wself.currItemModelIndex, section: 0)])
        }
        currItemModelIndex = max(0, currItemModelIndex - 1)
        
    }


    func endScrollItem(page: Int) {
        BFLog(1, message: "endScrollItem")
        // 切换素材时先把录制状态切为不可用,延迟可点，避免在缩略图未加载出来时即刻录制
        currItemModelIndex = page
        // 暂停
        pause()
        //add by ak  BFRecordScreenCameraManager 中的播放器也要暂停那里的进度回调也会刷新 UI 导致出现闪烁 BUG
        rscmanager.pause()
 
        // 如果在录制中，停止录制
        if isRecording {
            endRecord()
        }
        recorderManager?.cancelTitleService()
        
        events = [WithDrawModel]()
        
        
        if let recordItem = rscurrentManager.recordItem {
            // 暂停状态--如果是图片素材同时没有录音文件时不显示播放按钮
            playBtn.isSelected = (recordItem.mediaType != .Video && recordItem.voiceStickers.count <= 0)
            playBtn.isHidden = playBtn.isSelected
            // 重绘录音区域
//            progressThumV.recordItem = recordItem
            DispatchQueue.main.asyncAfter(deadline: .now() + 0.1) { [weak self] in
                guard let wself = self else { return }
                
                /// 重绘录音进度视图
                wself.resetAllIndirectionView()
                // 重绘录音进度
                wself.recordBtn.isEnabled = true
            }
            // 更新缩略图
            // progressThumV.isHidden = false
            
            // 切换要更新当前录制index，避免在录制完以后切换素材这种时候为-3会拦截
            currentPlayRecordIndex = -1
            showSubtitleIndex = -1
            // 重置播放器
            assetPlayer.seek(to: CMTime.zero)
            recordPlayer.seek(to: CMTime.zero)
            
            if let voice = itemModels[page].voiceStickers.enumerated().first(where: { m in
                m.1.startTime == 0
            }) {
                currentPlayRecordIndex = voice.0
            }
            
            searchStopAtRecordRange()
            changeWithDrawBtnLayout(0)
            pauseTime = 0
            
            if changeItemHandle != nil {
                changeItemHandle!(page)
            }
            
            // add by ak 切换后刷新字幕
            subtitleLabel.removeFromSuperview()
            let cell = collectionView.cellForItem(at: IndexPath(item: page, section: 0)) as? BFImageCoverViewCell
            cell?.contentView.addSubview(subtitleLabel)
            updateSubtitle(time: currentAssetProgress)
            
        }
    }

    public func updateContentOffset(index: Int) {
        let offx = collectionView.contentOffset.x
        if offx != collectionView.frame.width * CGFloat(index) {
            collectionView.setContentOffset(CGPoint(x: collectionView.frame.width * CGFloat(index), y: 0), animated: true)
        }
    }
}

// MARK: - 处理图片素材

/// 处理图片素材
public extension BFRecordScreenController {
    /// 播放录音
    /// - Parameter time: <#time description#>
    func startPlayRecord(time: CMTime) {
        // 播放对应的录音音频
        if currMediaType == .Image {
            if itemModels[currItemModelIndex].materialDuraion.seconds <= 0 {
                playBtn.isSelected = true
                playBtn.isHidden = true
                return
            }
            isNormalPlaying = true
            // 当开始播放时重置录音播放起始时间
            recordStartPlayTime = currentAssetProgress
            currenStartPlayTime = CMTime.zero
        }
        playRecord(at: time, periodicTimeObserver: { [weak self] currentT, _ in
            
            guard let wself = self else { return }
            
            BFLog(1, message: "播放录音进度：\(currentT.seconds)")
            if wself.currMediaType == .Image, wself.isNormalPlaying {
                wself.imageRecordProgress(progress: CMTimeGetSeconds(currentT))
            }
        }, didPlayToEndTime: { [weak self] recordInfo, currentItem in
            
            guard let wself = self else { return }
            
            BFLog(message: "播放录音结束：\(String(describing: recordInfo?.1)),\(String(describing: currentItem))")
            if wself.itemModels[wself.currItemModelIndex ].mediaType == .Image {
                if (wself.itemModels[wself.currItemModelIndex ].voiceStickers.count ) <= ((recordInfo?.0 ?? 0) + 1) || CMTimeCompare((recordInfo?.1.endCMTime ?? .zero), (wself.itemModels[wself.currItemModelIndex].voiceStickers.last?.endCMTime ?? .zero)) >= 0 {
                    
                    wself.isEndPlay = true
                    wself.pause()
                    // 注：矫正进度--播放结束后当前指针应该到当前素材总时长
                    wself.rscurrentManager.currentAssetProgress = CMTime(seconds: wself.itemModels[wself.currItemModelIndex ].materialDuraion.seconds , preferredTimescale: 1000)
                    wself.resetCurrentProgress()
                    // 录制播放结束后显示录制按钮
                    wself.recordBtn.isHidden = false
                } else {
                    // 注：矫正进度--一段录音播放结束后当前指针应该到当前录音结束点
                    let next = wself.itemModels[wself.currItemModelIndex ].voiceStickers[(recordInfo?.0 ?? 0) + 1].startCMTime
                    if CMTimeCompare(next, wself.currentAssetProgress) > 0{
                        wself.rscurrentManager.currentAssetProgress = next
                    }
                    wself.startPlayRecord(time: wself.currentAssetProgress )
                }
            }
        }) { [weak self] _, _ in
            
            guard let wself = self else { return }
            
            if wself.currMediaType == .Image {
                DispatchQueue.main.asyncAfter(deadline: DispatchTime.now() + 3) {[weak self] in
                    guard let wself = self else { return }
                    
                    wself.startPlayRecord(time: wself.currentAssetProgress)
                }
            }
        }
    }

    /// 处理图片素材录音
    func imageRecordProgress(isRecord: Bool = false, progress: Float64) {
        if isRecord {
            rscurrentManager.currentAssetProgress = (recorderManager?.voiceModel?.startCMTime ?? .zero) + CMTime(seconds: progress, preferredTimescale: 1000)
        } else {
            rscurrentManager.currentAssetProgress = recordStartPlayTime + CMTime(seconds: progress, preferredTimescale: 1000)
        }
//        BFLog(1, message: "图片录音进度：\(progress),currentAssetProgress=\(currentAssetProgress.seconds),\(itemModels[currItemModelIndex].materialDuraion)")
        if currMediaType == .Image {
            /// 重置进度
            resetCurrentProgress()
        }
    }

    /// 重置进度
    func resetCurrentProgress() {
        DispatchQueue.main.async {[weak self] in
            guard let wself = self else { return }
            
            if !(wself.isRecording), CMTimeCompare((wself.currentAssetProgress), (wself.itemModels[wself.currItemModelIndex].materialDuraion  )) > 0 {
                return
            }
//            BFLog(1, message: "更新录音进度\(#function)-\(wself.currentAssetProgress.seconds ?? 0)")
            wself.progreddL.text = String(format: "%@", (wself.currentAssetProgress.seconds).formatDurationToHMS())
            wself.progressThumV.progress = (wself.currentAssetProgress.seconds)
            wself.updateSubtitle(time: wself.currentAssetProgress)
        }
    }

    /// 重绘录音进度视图
    func resetAllIndirectionView() {
        if currMediaType == .Camera {
            rscmanager.cameraProgressV?.resetThumIV()
            rscmanager.prepareToPlayNext(needPlay: false)
            return
        }
        // 重绘录音进度视图
        var percenWidth: CGFloat = 0
        if currMediaType == .Image {
            percenWidth = progressThumV.thumbImageWidth / 2.0
        } else {
            percenWidth = progressThumV.progessIndicateBackV.frame.width / CGFloat(itemModels[currItemModelIndex].materialDuraion.seconds)
        }
        indirectionView?.resetAllSubViews(items: itemModels[currItemModelIndex].voiceStickers, percenWidth: percenWidth, totalDuration: itemModels[currItemModelIndex].materialDuraion.seconds)
    }
}
